# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2018 Gaël Écorchard <galou_breizh@yahoo.fr>
# SPDX-FileCopyrightText: 2022 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

"""Unified handler for FreeCAD macros that can be obtained from different sources."""
import base64
import binascii
import os
import re
import io
import codecs
from html import unescape
from typing import Dict, Tuple, List, Union, Optional

from addonmanager_macro_parser import MacroParser
import addonmanager_utilities as utils

import addonmanager_freecad_interface as fci

translate = fci.translate


#  @package AddonManager_macro
#  \ingroup ADDONMANAGER
#  \brief Unified handler for FreeCAD macros that can be obtained from
#  different sources
#  @{


class Macro:
    """This class provides a unified way to handle macros coming from different
    sources"""

    # Use a stored class variable for this so that we can override it during testing
    blocking_get = None

    # pylint: disable=too-many-instance-attributes
    def __init__(self, name):
        self.name = name
        self.on_wiki = False
        self.on_git = False
        self.desc = ""
        self.comment = ""
        self.code = ""
        self.url = ""
        self.raw_code_url = ""
        self.wiki = ""
        self.version = ""
        self.license = ""
        self.date = ""
        self.src_filename = ""
        self.filename_from_url = ""
        self.author = ""
        self.icon = ""  # This is the raw data as set in the macro's code: it is typically a URL
        self.icon_source_url = ""
        self.icon_data = None
        self.xpm = ""  # Possible alternate icon data
        self.other_files = []
        self.other_files_data = {}  # Base64-encoded data loaded from "other files"
        self.parsed = False
        self._console = fci.Console
        if Macro.blocking_get is None:
            Macro.blocking_get = utils.blocking_get

    def __eq__(self, other):
        return self.filename == other.filename

    @classmethod
    def from_cache(cls, cache_dict: Dict):
        """Use data from the cache dictionary to create a new macro, returning a
        reference to it."""
        instance = Macro(cache_dict["name"])
        for key, value in cache_dict.items():
            if key == "icon_data" and value is not None:
                try:
                    value = base64.b64decode(value)
                except binascii.Error as e:
                    fci.Console.PrintWarning(f"Failed to decode macro icon data: {e}\n")
            instance.__dict__[key] = value
        return instance

    def to_cache(self) -> Dict:
        """For cache purposes all public members of the class are returned. The binary icon data is
        base64-encoded."""
        cache_dict = {}
        for key, value in self.__dict__.items():
            if key == "icon_data" and value is not None:
                cache_dict[key] = base64.b64encode(value).decode("utf-8")
            elif key[0] != "_":
                cache_dict[key] = value
        return cache_dict

    @property
    def filename(self):
        """The filename of this macro"""
        if self.on_git:
            return os.path.basename(self.src_filename)
        elif self.filename_from_url:
            return self.filename_from_url
        return (self.name + ".FCMacro").replace(" ", "_")

    def is_installed(self):
        """Returns True if this macro is currently installed (that is, if it exists
        in the user macro directory), or False if it is not. Both the exact filename
        and the filename prefixed with "Macro", are considered an installation
        of this macro.
        """
        if self.on_git and not self.src_filename:
            return False
        return os.path.exists(
            os.path.join(fci.DataPaths().macro_dir, self.filename)
        ) or os.path.exists(os.path.join(fci.DataPaths().macro_dir, "Macro_" + self.filename))

    def fill_details_from_file(self, filename: str) -> None:
        """Opens the given Macro file and parses it for its metadata"""
        with open(filename, errors="replace", encoding="utf-8") as f:
            self.code = f.read()
            self.fill_details_from_code(self.code)
        if self.other_files:
            if self.on_git:
                self.load_other_files(os.path.dirname(filename) or ".")

    def fill_details_from_code(self, code: str) -> None:
        """Read the passed-in code and parse it for known metadata elements"""
        parser = MacroParser(self.name, code)
        for key, value in parser.parse_results.items():
            if value:
                self.__dict__[key] = value
        self.parsed = True

    def fill_details_from_wiki(self, url):
        """For a given URL, download its data and attempt to get the macro's metadata
        out of it. If the macro's code is hosted elsewhere, as specified by a
        "rawcodeurl" found on the wiki page, that code is downloaded and used as the
        source."""
        code = ""
        url = url.replace("freecadweb", "freecad")
        p = Macro.blocking_get(url)
        if not p:
            self._console.PrintWarning(
                translate(
                    "AddonsInstaller",
                    "Unable to open macro wiki page at {}",
                ).format(url)
                + "\n"
            )
            return
        p = p.decode("utf8")
        # check if the macro page has its code hosted elsewhere, download if
        # needed
        if "rawcodeurl" in p:
            code = self._fetch_raw_code(p)
        if not code:
            code = self._read_code_from_wiki(p)
        if not code:
            self._console.PrintWarning(
                translate("AddonsInstaller", "Unable to fetch the code of macro '{}'").format(
                    self.name
                )
                + "\n"
            )
            return

        desc = re.findall(
            r"<td class=\"ctEven left macro-description\">(.*?)</td>",
            p.replace("\n", " "),
        )
        if desc:
            desc = desc[0]
        else:
            self._console.PrintWarning(
                translate(
                    "AddonsInstaller",
                    "Unable to retrieve a description from the wiki for macro {}",
                ).format(self.name)
                + "\n"
            )
            desc = "No description available"
        self.desc = desc
        self.comment, _, _ = desc.partition("<br")  # Up to the first line break
        self.comment = re.sub(r"<.*?>", "", self.comment)  # Strip any tags
        self.url = url
        if isinstance(code, list):
            code = "".join(code)
        self.code = code
        self.fill_details_from_code(self.code)
        if not self.icon and not self.xpm and not self.icon_data:
            self.parse_wiki_page_for_icon(p)

        if not self.author:
            self.author = self.parse_desc("Author: ")
        if not self.date:
            self.date = self.parse_desc("Last modified: ")

    def _fetch_raw_code(self, page_data) -> Optional[str]:
        """Fetch code from the raw code URL specified on the wiki page."""
        code = None
        self.raw_code_url = re.findall(r'rawcodeurl.*?href="(http.*?)">', page_data)
        if self.raw_code_url:
            self.raw_code_url = self.raw_code_url[0]
            u2 = Macro.blocking_get(self.raw_code_url)
            if not u2:
                self._console.PrintWarning(
                    translate(
                        "AddonsInstaller",
                        "Unable to open macro code URL {}",
                    ).format(self.raw_code_url)
                    + "\n"
                )
                return None
            code = u2.decode("utf8")
            self._set_filename_from_url(self.raw_code_url)
        return code

    def _set_filename_from_url(self, url: str):
        lhs, slash, rhs = url.rpartition("/")
        if rhs.endswith(".py") or rhs.lower().endswith(".fcmacro"):
            self.filename_from_url = rhs

    @staticmethod
    def _read_code_from_wiki(p: str) -> Optional[str]:
        code = re.findall(r"<pre>(.*?)</pre>", p.replace("\n", "--endl--"))
        if code:
            # take the biggest code block
            code = str(sorted(code, key=len)[-1])
            code = code.replace("--endl--", "\n")
            # Clean HTML escape codes.
            code = unescape(code)
            code = code.replace(b"\xc2\xa0".decode("utf-8"), " ")
        return code

    def parse_desc(self, line_start: str) -> Union[str, None]:
        """Get data from the wiki for the value specified by line_start."""
        components = self.desc.split(">")
        for component in components:
            if component.startswith(line_start):
                end = component.find("<")
                return component[len(line_start) : end]
        return None

    def load_other_files(self, base_directory: str):
        """Loop over the list of "other files" and load them into our data store"""
        for other_file in self.other_files:
            if not other_file:
                continue
            if other_file == self.icon:
                self.other_files_data[other_file] = "ICON"  # Don't waste space storing it twice
                continue
            if other_file.startswith("/"):
                # Some macros had a leading slash even though they can't use an absolute path. Strip
                # it off.
                other_file = other_file[1:]
            src_file = os.path.normpath(os.path.join(base_directory, other_file))
            if not os.path.isfile(src_file):
                self._console.PrintWarning(f"Could not load {other_file} for macro {self.name}\n")
                continue
            with open(src_file, "rb") as f:
                self.other_files_data[other_file.replace(os.pathsep, "/")] = base64.b64encode(
                    f.read()
                ).decode("utf-8")

    def install(self, macro_dir: str) -> Tuple[bool, List[str]]:
        """Install a macro and all its related files
        Returns True if the macro was installed correctly.
        Parameters
        ----------
        - macro_dir: the directory to install into
        """

        if not self.code:
            return False, ["No code"]
        if not os.path.isdir(macro_dir):
            try:
                os.makedirs(macro_dir)
            except OSError:
                return False, [f"Failed to create {macro_dir}"]
        macro_path = os.path.join(macro_dir, self.filename)
        try:
            with codecs.open(macro_path, "w", "utf-8") as macrofile:
                macrofile.write(self.code)
        except OSError:
            return False, [f"Failed to write {macro_path}"]
        # Copy related files, which are supposed to be given relative to
        # self.src_filename.
        warnings = []

        self._copy_icon_data(macro_dir, warnings)
        success = self._copy_other_files(macro_dir, warnings)

        if warnings or not success > 0:
            return False, warnings

        self._console.PrintLog(f"Macro {self.name} was installed successfully.\n")
        return True, []

    def _copy_icon_data(self, macro_dir, warnings):
        """Copy any available icon data into the install directory"""
        base_dir = os.path.dirname(self.src_filename)
        if self.xpm:
            xpm_file = os.path.join(base_dir, self.name + "_icon.xpm")
            with open(xpm_file, "w", encoding="utf-8") as f:
                f.write(self.xpm)
        if self.icon and self.icon_data:
            filename = self.icon.rsplit("/", 1)[-1]
            try:
                with open(os.path.join(macro_dir, filename), "wb") as f:
                    f.write(self.icon_data)
            except (OSError, UnicodeDecodeError) as e:
                warnings.append(f"Failed to create {filename}")
                fci.Console.PrintWarning(f"Failed to create {filename}: {e}\n")

    def _copy_other_files(self, macro_dir, warnings) -> bool:
        """Copy any specified "other files" into the installation directory"""
        for filename, data in self.other_files_data.items():
            if not filename or not data or data == "ICON":
                continue
            filename = filename.replace("/", os.path.sep)
            full_path = os.path.join(macro_dir, filename)
            try:
                os.makedirs(os.path.dirname(full_path), exist_ok=True)
                with open(full_path, "wb") as f:
                    f.write(base64.b64decode(data))
            except (OSError, UnicodeDecodeError, binascii.Error) as e:
                warnings.append(f"Failed to create {filename}")
        return True  # No fatal errors, but some files may have failed to copy

    def _fetch_single_file(self, other_file, src_file, dst_file, warnings):
        if not os.path.isfile(src_file):
            # If the file does not exist, see if we have a raw code URL to fetch from
            if self.raw_code_url:
                fetch_url = self.raw_code_url.rsplit("/", 1)[0] + "/" + other_file
                self._console.PrintLog(f"Attempting to fetch {fetch_url}...\n")
                fetch_url = fetch_url.replace("freecadweb", "freecad")
                p = Macro.blocking_get(fetch_url)
                if p:
                    with open(dst_file, "wb") as f:
                        f.write(p)
                else:
                    self._console.PrintWarning(
                        translate(
                            "AddonsInstaller",
                            "Unable to fetch macro-specified file {} from {}",
                        ).format(other_file, fetch_url)
                        + "\n"
                    )
            else:
                warnings.append(
                    translate(
                        "AddonsInstaller",
                        "Could not locate macro-specified file {} (expected at {})",
                    ).format(other_file, src_file)
                )

    def parse_wiki_page_for_icon(self, page_data: str) -> None:
        """Attempt to find the url for the icon in the wiki page. Sets 'self.icon' if
        found."""

        # Method 1: the text "toolbar icon" appears on the page, and provides a direct
        # link to an icon

        # pylint: disable=line-too-long
        # Try to get an icon from the wiki page itself:
        # <a rel="nofollow" class="external text"
        # href="https://wiki.freecad.org/images/f/f5/blah.png">ToolBar Icon</a>
        icon_regex = re.compile(r'.*href="(.*?)">ToolBar Icon', re.IGNORECASE)
        wiki_icon = ""
        if "ToolBar Icon" in page_data:
            f = io.StringIO(page_data)
            lines = f.readlines()
            for line in lines:
                if ">ToolBar Icon<" in line:
                    match = icon_regex.match(line)
                    if match:
                        wiki_icon = match.group(1)
                        if "file:" not in wiki_icon.lower():
                            self.icon = wiki_icon
                            return
                        break

        # See if we found an icon, but it wasn't a direct link:
        icon_regex = re.compile(r'.*img.*?src="(.*?)"', re.IGNORECASE)
        if wiki_icon.startswith("http"):
            # It's a File: wiki link. We can load THAT page and get the image from it...
            self._console.PrintLog(f"Found a File: link for macro {self.name} -- {wiki_icon}\n")
            wiki_icon = wiki_icon.replace("freecadweb", "freecad")
            p = Macro.blocking_get(wiki_icon)
            if p:
                p = p.decode("utf8")
                f = io.StringIO(p)
                lines = f.readlines()
                trigger = False
                for line in lines:
                    if trigger:
                        match = icon_regex.match(line)
                        if match:
                            wiki_icon = match.group(1)
                            self.icon = "https://wiki.freecad.org/" + wiki_icon
                            return
                    elif "fullImageLink" in line:
                        trigger = True

            #    <div class="fullImageLink" id="file">
            #        <a href="/images/a/a2/Bevel.svg">
            #            <img alt="File:Bevel.svg" src="/images/a/a2/Bevel.svg"
            #            width="64" height="64"/>
            #        </a>


#  @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 FreeCAD Developers                                 *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to create copies of objects."""
## @package make_copy
# \ingroup draftmake
# \brief Provides functions to create copies of objects.

## \addtogroup draftmake
# @{
import FreeCAD as App
import draftutils.utils as utils
import draftutils.gui_utils as gui_utils


def make_copy(obj, force=None, reparent=False, simple_copy=False):
    """make_copy(object, [force], [reparent], [simple_copy])

    Make an exact copy of an object and return it.

    Parameters
    ----------
    obj :
        Object to copy.

    force :
        Obsolete, not used anymore.

    reparent :
        Group the new object in the same group of the old one.

    simple_copy :
        Create a simple copy of the object (a new non parametric
        Part::Feature with the same Shape as the given object).
    """
    if not App.ActiveDocument:
        App.Console.PrintError("No active document. Aborting\n")
        return

    newobj = None

    if simple_copy and hasattr(obj, "Shape"):
        # this was the old implementation that is actually not used by default
        _name = utils.get_real_name(obj.Name)
        newobj = App.ActiveDocument.addObject("Part::Feature", _name)
        newobj.Shape = obj.Shape
        gui_utils.format_object(newobj, obj)
    elif not simple_copy:
        # this is the new implementation using doc.copyObject API
        if obj.hasExtension("App::OriginGroupExtension"):
            # always copy with dependencies when copying App::Part and PartDesign::Body
            newobj = App.ActiveDocument.copyObject(obj, True)
        else:
            newobj = App.ActiveDocument.copyObject(obj)

    if not newobj:
        return None

    if reparent:
        parents = obj.InList
        if parents:
            for par in parents:
                if par.isDerivedFrom("App::DocumentObjectGroup") or par.isDerivedFrom("App::Part"):
                    par.addObject(newobj)
                else:
                    # That's the case of Arch_BuildingParts or Draft_Layers for example
                    if "Group" in par.PropertiesList:
                        if obj in par.Group:
                            group = par.Group
                            group.append(newobj)
                            par.Group = group

    return newobj


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later
# -*- coding: utf8 -*-

#***************************************************************************
#*   Copyright (c) 2012 Keith Sloan <keith@sloan-home.co.uk>               *
#*                                                                         *
#*   This program is free software; you can redistribute it and/or modify  *
#*   it under the terms of the GNU Lesser General Public License (LGPL)    *
#*   as published by the Free Software Foundation; either version 2 of     *
#*   the License, or (at your option) any later version.                   *
#*   for detail see the LICENCE text file.                                 *
#*                                                                         *
#*   This program is distributed in the hope that it will be useful,       *
#*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
#*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
#*   GNU Library General Public License for more details.                  *
#*                                                                         *
#*   You should have received a copy of the GNU Library General Public     *
#*   License along with this program; if not, write to the Free Software   *
#*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
#*   USA                                                                   *
#*                                                                         *
#*   Acknowledgements:                                                     *
#*                                                                         *
#*     Thanks to shoogen on the FreeCAD forum and Peter Li                 *
#*     for programming advice and some code.                               *
#*                                                                         *
#*                                                                         *
#***************************************************************************
__title__ = "FreeCAD OpenSCAD Workbench - CSG importer"
__author__ = "Keith Sloan <keith@sloan-home.co.uk>"
__url__ = ["http://www.sloan-home.co.uk/ImportCSG"]

printverbose = False

import io
import os

import xml.sax

import FreeCAD
import Part
import Draft

from OpenSCADFeatures import *
from OpenSCADUtils import *


import ply.lex as lex
import ply.yacc as yacc

params = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/OpenSCAD")
printverbose = params.GetBool('printVerbose', False)

if FreeCAD.GuiUp:
    gui = True
else:
    if printverbose: print("FreeCAD Gui not present.")
    gui = False

hassetcolor = []
alreadyhidden = []
original_root_objects = []

# Get the token map from the lexer. This is required.
import tokrules
from tokrules import tokens
from builtins import open as pyopen

translate = FreeCAD.Qt.translate


def shallHide(subject):
    for obj in subject.OutListRecursive:
        if "Matrix_Union" in str(obj.FullName):
            return False
        if "Extrude" in str(obj.FullName):
            return True
    return False


def setColorRecursively(obj, color, transp):
    '''
    For some reason a part made by cutting or fusing other parts do not have a color
    unless its constituents are also colored. This code sets colors for those
    constituents unless already set elsewhere.
    '''
    obj.ViewObject.ShapeColor = color
    obj.ViewObject.Transparency = transp
    # Add any other relevant features to this list
    boolean_features = ["Part::Fuse", "Part::MultiFuse", "Part::Cut",
                        "Part::Common", "Part::MultiCommon"]
    if obj.TypeId in boolean_features:
        for currentObject in obj.OutList:
            if printverbose: print(f"Fixing up colors for: {currentObject.FullName}")
            if currentObject not in hassetcolor:
                setColorRecursively(currentObject, color, transp)


def fixVisibility():
    # After an import, only the remaining root objects that we created should be visible, not any
    # of their individual component objects. But make sure to only handle the ones we just imported,
    # not anything that already existed. And objects that exist at the toplevel without any
    # children are ignored.
    for root_object in FreeCAD.ActiveDocument.RootObjects:
        if root_object not in original_root_objects:
            root_object.ViewObject.Visibility = True
            for obj in root_object.OutListRecursive:
                obj.ViewObject.Visibility = False


def open(filename):
    "called when freecad opens a file."
    global doc
    global pathName
    docname = os.path.splitext(os.path.basename(filename))[0]
    doc = FreeCAD.newDocument(docname)
    if filename.lower().endswith('.scad'):
        tmpfile = callopenscad(filename)
        pathName = os.path.dirname(os.path.normpath(filename))
        processcsg(tmpfile)
        try:
            os.unlink(tmpfile)
        except OSError:
            pass
    else:
        pathName = os.path.dirname(os.path.normpath(filename))
        processcsg(filename)
    return doc


def insert(filename, docname):
    "called when freecad imports a file"
    global doc
    global pathName
    groupname_unused = os.path.splitext(os.path.basename(filename))[0]
    try:
        doc = FreeCAD.getDocument(docname)
        for obj in doc.RootObjects:
            original_root_objects.append(obj)
    except NameError:
        doc = FreeCAD.newDocument(docname)
    if filename.lower().endswith('.scad'):
        tmpfile = callopenscad(filename)
        pathName = os.path.dirname(os.path.normpath(filename))
        processcsg(tmpfile)
        try:
            os.unlink(tmpfile)
        except OSError:
            pass
    else:
        pathName = os.path.dirname(os.path.normpath(filename))
        processcsg(filename)

def processcsg(filename):
    global doc

    if printverbose: print('ImportCSG Version 0.6a')
    # Build the lexer
    if printverbose: print('Start Lex')
    lex.lex(module=tokrules)
    if printverbose: print('End Lex')

    # Build the parser
    if printverbose: print('Load Parser')
    # Disable generation of debug ('parser.out') and table cache ('parsetab.py'),
    # as it requires a writable location
    parser = yacc.yacc(debug=False, write_tables=False)
    if printverbose: print('Parser Loaded')

    f = io.open(filename, 'r', encoding="utf8")

    if printverbose: print('Start Parser')

    result = parser.parse(f.read())
    f.close()
    if printverbose:
        print('End Parser')
        print(result)
    if gui:
        fixVisibility()
    hassetcolor.clear()
    alreadyhidden.clear()
    FreeCAD.Console.PrintMessage('End processing CSG file\n')
    doc.recompute()


def p_block_list_(p):
    '''
    block_list : statement
               | block_list statement
               | statementwithmod
               | block_list statementwithmod
    '''

    if(len(p) > 2):
        if printverbose: print(p[2])
        p[0] = p[1] + p[2]
    else:
        p[0] = p[1]

def p_render_action(p):
    'render_action : render LPAREN keywordargument_list RPAREN OBRACE block_list EBRACE'
    if printverbose: print("Render (ignored)")
    p[0] = p[6]


def p_group_action1(p):
    'group_action1 : group LPAREN RPAREN OBRACE block_list EBRACE'
    if printverbose: print("Group")
    # Test need for implicit fuse
    if p[5] is None:
        p[0] = []
        return
    if len(p[5]) > 1:
        if printverbose: print('Fuse Group')
        for obj in p[5]:
            checkObjShape(obj)
        p[0] = [fuse(p[5], "Group")]
    else:
        if printverbose: print(f"Group {p[5]} type {type(p[5])}")
        checkObjShape(p[5])
        p[0] = p[5]


def p_group_action2(p):
    'group_action2 : group LPAREN RPAREN SEMICOL'
    if printverbose: print("Group2")
    p[0] = []


def p_boolean(p):
    '''
    boolean : true
            | false
    '''
    p[0] = p[1]


def p_stripped_string(p):
    'stripped_string : STRING'
    p[0] = p[1].strip('"')


def p_statement(p):
    '''statement : part
                 | operation
                 | multmatrix_action
                 | group_action1
                 | group_action2
                 | color_action
                 | render_action
                 | not_supported
    '''
    p[0] = p[1]


def p_anymodifier(p):
    '''anymodifier : MODIFIERBACK
                   | MODIFIERDEBUG
                   | MODIFIERROOT
                   | MODIFIERDISABLE
    '''
    # just return the plain modifier for now
    # has to be changed when the modifiers are implemented
    # please note that disabled objects usually are stripped of the CSG output during compilation
    p[0] = p[1]


def p_statementwithmod(p):
    '''statementwithmod : anymodifier statement'''
    # ignore the modifiers but add them to the label
    modifier = p[1]
    obj = p[2]
    if hasattr(obj, 'Label'):
        obj.Label = modifier + obj.Label
    p[0] = obj


def p_part(p):
    '''
    part : sphere_action
         | cylinder_action
         | cube_action
         | circle_action
         | square_action
         | text_action
         | polygon_action_nopath
         | polygon_action_plus_path
         | polyhedron_action
         '''
    p[0] = p[1]


def p_2d_point(p):
    '2d_point : OSQUARE NUMBER COMMA NUMBER ESQUARE'
    global points_list
    if printverbose: print("2D Point")
    p[0] = [float(p[2]), float(p[4])]


def p_points_list_2d(p):
    '''
    points_list_2d : 2d_point COMMA
                   | points_list_2d 2d_point COMMA
                   | points_list_2d 2d_point
                   '''
    if p[2] == ',':
        p[0] = [p[1]]
    else:
        if printverbose:
            print(p[1])
            print(p[2])
        p[1].append(p[2])
        p[0] = p[1]


def p_3d_point(p):
    '3d_point : OSQUARE NUMBER COMMA NUMBER COMMA NUMBER ESQUARE'
    global points_list
    if printverbose: print("3d point")
    p[0] = [p[2], p[4], p[6]]


def p_points_list_3d(p):
    '''
    points_list_3d : 3d_point COMMA
               | points_list_3d 3d_point COMMA
               | points_list_3d 3d_point
               '''
    if p[2] == ',':
        if printverbose: print("Start List")
        if printverbose: print(p[1])
        p[0] = [p[1]]
    else:
        if printverbose: print(p[1])
        if printverbose: print(p[2])
        p[1].append(p[2])
        p[0] = p[1]
    if printverbose: print(p[0])

def p_path_points(p):
    '''
    path_points : NUMBER COMMA
                | path_points NUMBER COMMA
                | path_points NUMBER
                '''

    if p[2] == ',':
        p[0] = [int(p[1])]
    else:
        p[1].append(int(p[2]))
        p[0] = p[1]


def p_path_list(p):
    'path_list : OSQUARE path_points ESQUARE'
    p[0] = p[2]


def p_path_set(p):
    '''
    path_set : path_list
             | path_set COMMA path_list
             '''
    if len(p) == 2:
        p[0] = [p[1]]
    else:
        p[1].append(p[3])
        p[0] = p[1]

def p_operation(p):
    '''
    operation : difference_action
              | intersection_action
              | union_action
              | rotate_extrude_action
              | linear_extrude_with_transform
              | rotate_extrude_file
              | import_file1
              | resize_action
              | surface_action
              | projection_action
              | hull_action
              | minkowski_action
              | offset_action
              '''
    p[0] = p[1]

def placeholder(name, children, arguments):
    from OpenSCADFeatures import OpenSCADPlaceholder
    newobj=doc.addObject("Part::FeaturePython",name)
    OpenSCADPlaceholder(newobj, children, str(arguments))
    if gui:
        if FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/OpenSCAD").\
            GetBool('useViewProviderTree'):
            from OpenSCADFeatures import ViewProviderTree
            ViewProviderTree(newobj.ViewObject)
        else:
            newobj.ViewObject.Proxy = 0
    #don't hide the children
    return newobj

def CGALFeatureObj(name, children,arguments=[]):
    myobj=doc.addObject("Part::FeaturePython", name)
    CGALFeature(myobj, name, children, str(arguments))
    if gui:
        for subobj in children:
            subobj.ViewObject.hide()
        if FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/OpenSCAD").\
            GetBool('useViewProviderTree'):
            from OpenSCADFeatures import ViewProviderTree
            ViewProviderTree(myobj.ViewObject)
        else:
            myobj.ViewObject.Proxy = 0
    return myobj

def p_offset_action(p):
    'offset_action : offset LPAREN keywordargument_list RPAREN OBRACE block_list EBRACE'
    subobj = None
    if len(p[6]) == 0:
        newobj = placeholder('group',[],'{}')
    elif (len(p[6]) == 1 ): #single object
        subobj = p[6][0]
    else:
        subobj = fuse(p[6],"Offset Union")
    if 'r' in p[3]:
        offset = float(p[3]['r'])
    if 'delta' in p[3]:
        offset = float(p[3]['delta'])
    checkObjShape(subobj)
    if subobj.Shape.Volume == 0 :
        newobj = doc.addObject("Part::Offset2D",'Offset2D')
        newobj.Source = subobj
        newobj.Value = offset
        if 'r' in p[3]:
            newobj.Join = 0
        else:
            newobj.Join = 2
    else:
        newobj = doc.addObject("Part::Offset",'offset')
        newobj.Shape = subobj[0].Shape.makeOffset(offset)
    newobj.Document.recompute()
    if gui:
        subobj.ViewObject.hide()
    p[0] = [newobj]

def checkObjShape(obj):
    if printverbose: print('Check Object Shape')
    if hasattr(obj, 'Shape'):
        if obj.Shape.isNull():
            if printverbose: print('Shape is Null - recompute')
            obj.recompute()
        if obj.Shape.isNull():
            print(f'Recompute failed : {obj.Name}')
    else:
        if hasattr(obj, 'Name'):
            print(f"obj {obj.Name} has no Shape")
        else:
            print(f"obj {obj} has no Name & Shape")

def p_hull_action(p):
    'hull_action : hull LPAREN RPAREN OBRACE block_list EBRACE'
    p[0] = [ CGALFeatureObj(p[1],p[5]) ]

def p_minkowski_action(p):
    '''
    minkowski_action : minkowski LPAREN keywordargument_list RPAREN OBRACE block_list EBRACE'''
    p[0] = [ CGALFeatureObj(p[1],p[6],p[3]) ]

def p_resize_action(p):
    '''
    resize_action : resize LPAREN keywordargument_list RPAREN OBRACE block_list EBRACE '''
    new_size = p[3]['newsize']
    auto     = p[3]['auto']
    p[6][0].recompute()
    if p[6][0].Shape.isNull():
        doc.recompute()
    p[6][0].Shape.tessellate(0.05)
    old_bbox = p[6][0].Shape.BoundBox
    old_size = [old_bbox.XLength, old_bbox.YLength, old_bbox.ZLength]
    for r in range(0,3):
        if auto[r] == '1':
            new_size[r] = new_size[0]
        if new_size[r] == '0':
            new_size[r] = str(old_size[r])

    # Calculate a transform matrix from the current bounding box to the new one:
    transform_matrix = FreeCAD.Matrix()

    scale = FreeCAD.Vector(float(new_size[0])/old_size[0],
                           float(new_size[1])/old_size[1],
                           float(new_size[2])/old_size[2])

    transform_matrix.scale(scale)

    new_part=doc.addObject("Part::FeaturePython",'Matrix Deformation')
    new_part.Shape = p[6][0].Shape.transformGeometry(transform_matrix)
    if gui:
        if FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/OpenSCAD").\
            GetBool('useViewProviderTree'):
            from OpenSCADFeatures import ViewProviderTree
            ViewProviderTree(new_part.ViewObject)
        else:
            new_part.ViewObject.Proxy = 0
        p[6][0].ViewObject.hide()
    p[0] = [new_part]


def p_not_supported(p):
    '''
    not_supported : glide LPAREN keywordargument_list RPAREN OBRACE block_list EBRACE
                  | subdiv LPAREN keywordargument_list RPAREN OBRACE block_list EBRACE
                  '''
    if gui and not FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/OpenSCAD").\
            GetBool('usePlaceholderForUnsupported'):
        from PySide import QtGui
        QtGui.QMessageBox.critical(None, translate('OpenSCAD',"Unsupported Function")+" : "+p[1],translate('OpenSCAD',"Press OK"))
    else:
        p[0] = [placeholder(p[1],p[6],p[3])]

def p_size_vector(p):
    'size_vector : OSQUARE NUMBER COMMA NUMBER COMMA NUMBER ESQUARE'
    if printverbose: print("size vector")
    p[0] = [p[2],p[4],p[6]]

def p_keywordargument(p):
    '''keywordargument : ID EQ boolean
    | ID EQ NUMBER
    | ID EQ size_vector
    | ID EQ vector
    | ID EQ 2d_point
    | text EQ stripped_string
    | ID EQ stripped_string
     '''
    p[0] = (p[1],p[3])
    if printverbose: print(p[0])

def p_keywordargument_list(p):
    '''
    keywordargument_list : keywordargument
               | keywordargument_list COMMA keywordargument
    '''
    if len(p) == 2:
        p[0] = {p[1][0] : p[1][1]}
    else:
        p[1][p[3][0]] = p[3][1]
        p[0]=p[1]

def p_color_action(p):
    'color_action : color LPAREN vector RPAREN OBRACE block_list EBRACE'
    import math
    if printverbose: print("Color")
    color = tuple([float(f) for f in p[3][:3]]) #RGB
    transp = 100 - int(math.floor(100*float(p[3][3]))) #Alpha
    if gui:
        for obj in p[6]:
            if shallHide(obj):
                if "Group" in obj.FullName:
                    obj.ViewObject.Visibility=False
                    alreadyhidden.append(obj)
            setColorRecursively(obj, color, transp)
            hassetcolor.append(obj)
    p[0] = p[6]

# Error rule for syntax errors
def p_error(p):
    if printverbose: print("Syntax error in input!")
    if printverbose: print(p)

def fuse(lst,name):
    global doc
    if printverbose: print("Fuse")
    if printverbose: print(lst)
    if len(lst) == 0:
        myfuse = placeholder('group',[],'{}')
    elif len(lst) == 1:
        return lst[0]
    # Is this Multi Fuse
    elif len(lst) > 2:
        if printverbose: print("Multi Fuse")
        myfuse = doc.addObject('Part::MultiFuse',name)
        myfuse.Shapes = lst
        if gui:
            for subobj in myfuse.Shapes:
                subobj.ViewObject.hide()
    else:
        if printverbose: print("Single Fuse")
        myfuse = doc.addObject('Part::Fuse',name)
        myfuse.Base = lst[0]
        myfuse.Tool = lst[1]
        checkObjShape(myfuse.Base)
        checkObjShape(myfuse.Tool)
        myfuse.Shape = myfuse.Base.Shape.fuse(myfuse.Tool.Shape)
        if gui:
            myfuse.Base.ViewObject.hide()
            myfuse.Tool.ViewObject.hide()
    myfuse.Placement = FreeCAD.Placement()
    return myfuse

def p_empty_union_action(p):
    'union_action : union LPAREN RPAREN SEMICOL'
    if printverbose: print("empty union")
    newpart = fuse([],p[1])
    if printverbose: print("Push Union Result")
    p[0] = [newpart]
    if printverbose: print("End Union")

def p_union_action(p):
    'union_action : union LPAREN RPAREN OBRACE block_list EBRACE'
    if printverbose: print("union")
    newpart = fuse(p[5],p[1])
    if printverbose: print("Push Union Result")
    p[0] = [newpart]
    if printverbose: print("End Union")

def p_difference_action(p):
    'difference_action : difference LPAREN RPAREN OBRACE block_list EBRACE'

    if printverbose: print("difference")
    if printverbose: print(len(p[5]))
    if printverbose: print(p[5])
    if (len(p[5]) == 0 ): #nochild
        mycut_unused = placeholder('group',[],'{}')
    elif (len(p[5]) == 1 ): #single object
        p[0] = p[5]
    else:
# Cut using Fuse
        mycut = doc.addObject('Part::Cut',p[1])
        mycut.Base = p[5][0]
#       Can only Cut two objects do we need to fuse extras
        if (len(p[5]) > 2 ):
            if printverbose: print("Need to Fuse Extra First")
            mycut.Tool = fuse(p[5][1:],'union')
        else :
            mycut.Tool = p[5][1]
            checkObjShape(mycut.Tool)
        if gui:
            mycut.Base.ViewObject.hide()
            mycut.Tool.ViewObject.hide()
        if printverbose: print("Push Resulting Cut")
        p[0] = [mycut]
    if printverbose: print("End Cut")

def p_intersection_action(p):
    'intersection_action : intersection LPAREN RPAREN OBRACE block_list EBRACE'

    if printverbose: print("intersection")
    # Is this Multi Common
    if (len(p[5]) > 2):
        if printverbose: print("Multi Common")
        mycommon = doc.addObject('Part::MultiCommon',p[1])
        mycommon.Shapes = p[5]
        if gui:
            for subobj in mycommon.Shapes:
                subobj.ViewObject.hide()
    elif (len(p[5]) == 2):
        if printverbose: print("Single Common")
        mycommon = doc.addObject('Part::Common',p[1])
        mycommon.Base = p[5][0]
        mycommon.Tool = p[5][1]
        checkObjShape(mycommon.Base)
        checkObjShape(mycommon.Tool)
        if gui:
            mycommon.Base.ViewObject.hide()
            mycommon.Tool.ViewObject.hide()
    elif (len(p[5]) == 1):
        mycommon = p[5][0]
    else : # 1 child
        mycommon = placeholder('group',[],'{}')
    mycommon.Shape = mycommon.Base.Shape.common(mycommon.Tool.Shape)
    p[0] = [mycommon]
    if printverbose: print("End Intersection")

def process_rotate_extrude(obj, angle):
    newobj=doc.addObject("Part::FeaturePython",'RefineRotateExtrude')
    RefineShape(newobj,obj)
    if gui:
        if FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/OpenSCAD").\
            GetBool('useViewProviderTree'):
            from OpenSCADFeatures import ViewProviderTree
            ViewProviderTree(newobj.ViewObject)
        else:
            newobj.ViewObject.Proxy = 0
        obj.ViewObject.hide()
    myrev = doc.addObject("Part::Revolution","RotateExtrude")
    myrev.Source = newobj
    myrev.Axis = (0.00,1.00,0.00)
    myrev.Base = (0.00,0.00,0.00)
    myrev.Angle = angle
    myrev.Placement = FreeCAD.Placement(FreeCAD.Vector(),FreeCAD.Rotation(0,0,90))
    if gui:
        newobj.ViewObject.hide()
    return myrev

def process_rotate_extrude_prism(obj, angle, n):
    newobj=doc.addObject("Part::FeaturePython",'PrismaticToroid')
    PrismaticToroid(newobj, obj, angle, n)
    newobj.Placement=FreeCAD.Placement(FreeCAD.Vector(),FreeCAD.Rotation(0,0,90))
    if gui:
        if FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/OpenSCAD").\
            GetBool('useViewProviderTree'):
            from OpenSCADFeatures import ViewProviderTree
            ViewProviderTree(newobj.ViewObject)
        else:
            newobj.ViewObject.Proxy = 0
        obj.ViewObject.hide()
    return newobj

def p_rotate_extrude_action(p):
    'rotate_extrude_action : rotate_extrude LPAREN keywordargument_list RPAREN OBRACE block_list EBRACE'
    if printverbose: print("Rotate Extrude")
    angle = 360.0
    if 'angle' in p[3]:
        angle = float(p[3]['angle'])
    n = int(round(float(p[3]['$fn'])))
    fnmax = FreeCAD.ParamGet(\
        "User parameter:BaseApp/Preferences/Mod/OpenSCAD").\
        GetInt('useMaxFN', 16)
    if (len(p[6]) > 1) :
        part = fuse(p[6],"Rotate Extrude Union")
    else :
        part = p[6][0]

    if n < 3 or fnmax != 0 and n > fnmax:
        p[0] = [process_rotate_extrude(part,angle)]
    else:
        p[0] = [process_rotate_extrude_prism(part,angle,n)]
    if printverbose: print("End Rotate Extrude")

def p_rotate_extrude_file(p):
    'rotate_extrude_file : rotate_extrude LPAREN keywordargument_list RPAREN SEMICOL'
    if printverbose: print("Rotate Extrude File")
    angle = 360.0
    if 'angle' in p[3]:
        angle = float(p[3]['angle'])
    filen,ext = p[3]['file'] .rsplit('.',1)
    obj = process_import_file(filen,ext,p[3]['layer'])
    n = int(round(float(p[3]['$fn'])))
    fnmax = FreeCAD.ParamGet(\
        "User parameter:BaseApp/Preferences/Mod/OpenSCAD").\
        GetInt('useMaxFN', 16)

    if n < 3 or fnmax != 0 and n > fnmax:
        p[0] = [process_rotate_extrude(obj,angle)]
    else:
        p[0] = [process_rotate_extrude_prism(obj,angle,n)]
    if printverbose: print("End Rotate Extrude File")

def process_linear_extrude(obj,h) :
    #if gui:
    newobj = doc.addObject("Part::FeaturePython",'RefineLinearExtrude')
    RefineShape(newobj,obj)#mylinear)
    if gui:
        if FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/OpenSCAD").\
            GetBool('useViewProviderTree'):
            from OpenSCADFeatures import ViewProviderTree
            ViewProviderTree(newobj.ViewObject)
        else:
            newobj.ViewObject.Proxy = 0
        obj.ViewObject.hide()
        #mylinear.ViewObject.hide()
    mylinear = doc.addObject("Part::Extrusion","LinearExtrude")
    mylinear.Base = newobj #obj
    mylinear.Dir = (0,0,h)
    mylinear.Placement=FreeCAD.Placement()
    # V17 change to False mylinear.Solid = True
    mylinear.Solid = False
    if gui:
        newobj.ViewObject.hide()
    return mylinear

def process_linear_extrude_with_transform(base,height,twist,scale) :
    newobj = doc.addObject("Part::FeaturePython",'transform_extrude')
    # base is a FreeCAD Object, height & twist are floats, scale is a two-component vector of floats
    Twist(newobj,base,height,-twist,scale)
    if gui:
        if FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/OpenSCAD").\
            GetBool('useViewProviderTree'):
            from OpenSCADFeatures import ViewProviderTree
            ViewProviderTree(newobj.ViewObject)
        else:
            newobj.ViewObject.Proxy = 0
    return newobj

def p_linear_extrude_with_transform(p):
    'linear_extrude_with_transform : linear_extrude LPAREN keywordargument_list RPAREN OBRACE block_list EBRACE'
    if printverbose: print("Linear Extrude With Transform")
    h = float(p[3]['height'])
    if printverbose: print("Height : ",h)
    s = [1.0,1.0]
    t = 0.0
    if 'scale' in p[3]:
        if isinstance(p[3]['scale'], str):
            s = [float(p[3]['scale']), float(p[3]['scale'])]
        else:
            s = [float(p[3]['scale'][0]), float(p[3]['scale'][1])]
        if printverbose: print ("Scale: " + str(s))
    if 'twist' in p[3]:
        t = float(p[3]['twist'])
        if printverbose: print("Twist : ",t)
    # Test if null object like from null text
    if (len(p[6]) == 0) :
        p[0] = []
        return
    if (len(p[6]) > 1) :
        obj = fuse(p[6],"Linear Extrude Union")
    else :
        obj = p[6][0]
    checkObjShape(obj)
    if t != 0.0 or s[0] != 1.0 or s[1] != 1.0:
        newobj = process_linear_extrude_with_transform(obj,h,t,s)
    else:
        newobj = process_linear_extrude(obj,h)
    if p[3].get('center','false')=='true' :
        center(newobj,0,0,h)
    p[0] = [newobj]
    if gui:
        obj.ViewObject.hide()
    if printverbose: print("End Linear Extrude with Transform")

def p_import_file1(p):
    'import_file1 : import LPAREN keywordargument_list RPAREN SEMICOL'
    if printverbose: print("Import File")
    filen,ext = p[3]['file'].rsplit('.',1)
    p[0] = [process_import_file(filen,ext,p[3]['layer'])]
    if printverbose: print("End Import File")

def p_surface_action(p):
    'surface_action : surface LPAREN keywordargument_list RPAREN SEMICOL'
    if printverbose: print("Surface")
    obj = doc.addObject("Part::Feature",'surface')
    obj.Shape,xoff,yoff=makeSurfaceVolume(p[3]['file'])
    if p[3].get('center','false') == 'true' :
        center(obj,xoff,yoff,0.0)
    p[0] = [obj]
    if printverbose: print("End surface")

def process_import_file(fname,ext,layer):
    if printverbose: print("Importing : "+fname+"."+ext+" Layer : "+layer)
    if ext.lower() in reverseimporttypes()['Mesh']:
        obj = process_mesh_file(fname,ext)
    elif ext.lower() == 'dxf' :
        obj = processDXF(fname,layer)
    elif ext.lower() == 'svg':
        obj = processSVG(fname, ext)
    else:
        raise ValueError("Unsupported file extension %s" % ext)
    return obj

def processSVG(fname, ext):
    from importSVG import svgHandler
    if printverbose: print("SVG Handler")
    doc = FreeCAD.ActiveDocument
    docSVG = FreeCAD.newDocument(fname+'_tmp')
    FreeCAD.ActiveDocument = docSVG

    # Set up the parser
    parser = xml.sax.make_parser()
    parser.setFeature(xml.sax.handler.feature_external_ges, False)
    parser.setContentHandler(svgHandler())
    parser._cont_handler.doc = docSVG

    # pathName is a Global
    filename = os.path.join(pathName,fname+'.'+ext)
    # Use the native Python open which was saved as `pyopen`
    parser.parse(pyopen(filename))

    #combine SVG objects into one
    shapes = []
    for obj in FreeCAD.ActiveDocument.Objects:
        if printverbose: print(obj.Name)
        if printverbose: print(obj.Shape)
        shapes.append(obj.Shape)
    FreeCAD.closeDocument(docSVG.Name)
    FreeCAD.ActiveDocument=doc
    obj=doc.addObject('Part::Feature',fname)
    obj.Shape=Part.Compound(shapes)
    return obj

def process_mesh_file(fname,ext):
    import Mesh
    import Part
    fullname = fname+'.'+ext
    filename = os.path.join(pathName,fullname)
    objname = os.path.split(fname)[1]
    mesh1 = doc.getObject(objname) #reuse imported object
    if not mesh1:
        Mesh.insert(filename)
        mesh1 = doc.getObject(objname)
    if mesh1 is not None:
        if gui:
            mesh1.ViewObject.hide()
        sh = Part.Shape()
        sh.makeShapeFromMesh(mesh1.Mesh.Topology,0.1)
        solid = Part.Solid(sh)
        obj = doc.addObject('Part::Feature',"Mesh")

        solid = solid.removeSplitter()
        if solid.Volume < 0:
            solid.complement()
        obj.Shape = solid
    else: #mesh1 is None
        FreeCAD.Console.PrintError('Mesh not imported %s.%s %s\n' % \
                (objname,ext,filename))
        import Part
        obj = doc.addObject('Part::Feature',"FailedMeshImport")
        obj.Shape = Part.Compound([])
    return obj


def processTextCmd(t):
    from OpenSCADUtils import callopenscadstring
    tmpfilename = callopenscadstring(t,'dxf')
    from OpenSCAD2Dgeom import importDXFface
    face = importDXFface(tmpfilename,None,None)
    obj = doc.addObject('Part::Feature','text')
    obj.Shape = face
    try:
        os.unlink(tmpfilename)
    except OSError:
        pass
    return obj

def processDXF(fname,layer):
    global doc
    global pathName
    from OpenSCAD2Dgeom import importDXFface
    if printverbose: print("Process DXF file")
    if printverbose: print("File Name : "+fname)
    if printverbose: print("Layer : "+layer)
    if printverbose: print("PathName : "+pathName)
    dxfname = fname+'.dxf'
    filename = os.path.join(pathName,dxfname)
    shortname = os.path.split(fname)[1]
    if printverbose: print("DXF Full path : "+filename)
    face = importDXFface(filename,layer,doc)
    obj=doc.addObject('Part::Feature','dxf_%s_%s' % (shortname,layer or "all"))
    obj.Shape=face
    if printverbose: print("DXF Diagnostics")
    if printverbose: print(obj.Shape.ShapeType)
    if printverbose: print("Closed : "+str(obj.Shape.isClosed()))
    if printverbose: print(obj.Shape.check())
    if printverbose: print([w.isClosed() for w in obj.Shape.Wires])
    return obj

def processSTL(fname):
    if printverbose: print("Process STL file")

def p_multmatrix_action(p):
    'multmatrix_action : multmatrix LPAREN matrix RPAREN OBRACE block_list EBRACE'
    if printverbose: print("MultMatrix")
    transform_matrix = FreeCAD.Matrix()
    if printverbose: print("Multmatrix")
    if printverbose: print(p[3])
    if gui and p[6]:
        parentcolor=p[6][0].ViewObject.ShapeColor
        parenttransparency=p[6][0].ViewObject.Transparency

    m1l=sum(p[3],[])
    if any('x' in me for me in m1l): #hexfloats
        m1l=[float.fromhex(me) for me in m1l]
        matrixisrounded=False
    elif max((len(me) for me in m1l)) >= 14: #might have double precision
        m1l=[float(me) for me in m1l] # assume precise output
        m1l=[(0 if (abs(me) < 1e-15) else me) for me in m1l]
        matrixisrounded=False
    else: #trucanted numbers
        m1l=[round(float(me),12) for me in m1l] #round
        matrixisrounded=True
    transform_matrix = FreeCAD.Matrix(*tuple(m1l))
    if printverbose: print(transform_matrix)
    if printverbose: print("Apply Multmatrix")
#   If more than one object on the stack for multmatrix fuse first
    if (len(p[6]) == 0) :
        part = placeholder('group',[],'{}')
    elif (len(p[6]) > 1) :
        part = fuse(p[6],"Matrix Union")
    else :
        part = p[6][0]
    if (isspecialorthogonalpython(fcsubmatrix(transform_matrix))) :
        if printverbose: print("special orthogonal")
        if matrixisrounded:
            if printverbose: print("rotation rounded")
            plm=FreeCAD.Placement(transform_matrix)
            plm=FreeCAD.Placement(plm.Base,roundrotation(plm.Rotation))
            part.Placement=plm.multiply(part.Placement)
        else:
            part.Placement=FreeCAD.Placement(transform_matrix).multiply(\
                    part.Placement)
        new_part = part
    elif isrotoinversionpython(fcsubmatrix(transform_matrix)):
        if printverbose: print("orthogonal and inversion")
        cmat,axisvec = decomposerotoinversion(transform_matrix)
        new_part=doc.addObject("Part::Mirroring",'mirr_%s'%part.Name)
        new_part.Source=part
        new_part.Normal=axisvec
        if matrixisrounded:
            if printverbose: print("rotation rounded")
            plm=FreeCAD.Placement(cmat)
            new_part.Placement=FreeCAD.Placement(plm.Base,roundrotation(plm.Rotation))
        else:
            new_part.Placement=FreeCAD.Placement(cmat)
        new_part.Label="mirrored %s" % part.Label
        if gui:
            part.ViewObject.hide()
    elif FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/OpenSCAD").\
        GetBool('useMultmatrixFeature'):
        from OpenSCADFeatures import MatrixTransform
        new_part=doc.addObject("Part::FeaturePython",'Matrix Deformation')
        MatrixTransform(new_part,transform_matrix,part)
        if gui:
            if FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/OpenSCAD").\
                GetBool('useViewProviderTree'):
                from OpenSCADFeatures import ViewProviderTree
                ViewProviderTree(new_part.ViewObject)
            else:
                new_part.ViewObject.Proxy = 0
            part.ViewObject.hide()
    else :
        if printverbose: print("Transform Geometry")
        part.recompute()
        if part.Shape.isNull():
            doc.recompute()
        new_part = doc.addObject("Part::Feature","Matrix Deformation")
        new_part.Shape = part.Shape.transformGeometry(transform_matrix)
        if gui:
            part.ViewObject.hide()
    if False :
#   Does not fix problemfile or beltTighener although later is closer
        newobj=doc.addObject("Part::FeaturePython",'RefineMultMatrix')
        RefineShape(newobj,new_part)
        if gui:
            newobj.ViewObject.Proxy = 0
            new_part.ViewObject.hide()
        p[0] = [newobj]
    else :
        p[0] = [new_part]
    if gui and p[6]:
        new_part.ViewObject.ShapeColor=parentcolor
        new_part.ViewObject.Transparency = parenttransparency
    if printverbose: print("Multmatrix applied")

def p_matrix(p):
    'matrix : OSQUARE vector COMMA vector COMMA vector COMMA vector ESQUARE'
    if printverbose: print("Matrix")
    p[0] = [p[2],p[4],p[6],p[8]]

def p_vector(p):
    'vector : OSQUARE NUMBER COMMA NUMBER COMMA NUMBER COMMA NUMBER ESQUARE'
    if printverbose: print("Vector")
    p[0] = [p[2],p[4],p[6],p[8]]

def center(obj,x,y,z):
    obj.Placement = FreeCAD.Placement(\
        FreeCAD.Vector(-x/2.0,-y/2.0,-z/2.0),\
        FreeCAD.Rotation(0,0,0,1))

def p_sphere_action(p):
    'sphere_action : sphere LPAREN keywordargument_list RPAREN SEMICOL'
    if printverbose: print("Sphere : ",p[3])
    r = float(p[3]['r'])
    mysphere = doc.addObject("Part::Sphere",p[1])
    mysphere.Radius = r
    if printverbose: print("Push Sphere")
    p[0] = [mysphere]
    if printverbose: print("End Sphere")

def myPolygon(n,r1):
    # Adapted from Draft::_Polygon
    import math
    if printverbose: print("My Polygon")
    angle = math.pi*2/n
    nodes = [FreeCAD.Vector(r1,0,0)]
    for i in range(n-1) :
        th = (i+1) * angle
        nodes.append(FreeCAD.Vector(r1*math.cos(th),r1*math.sin(th),0))
    nodes.append(nodes[0])
    polygonwire = Part.makePolygon(nodes)

    polygon = doc.addObject("Part::Feature","Polygon")
    polygon.Shape = Part.Face(polygonwire)
    return polygon

def p_cylinder_action(p):
    'cylinder_action : cylinder LPAREN keywordargument_list RPAREN SEMICOL'
    if printverbose: print("Cylinder")
    tocenter = p[3].get('center','false')
    h = float(p[3]['h'])
    r1 = float(p[3]['r1'])
    r2 = float(p[3]['r2'])
    n = int(round(float(p[3]['$fn'])))
    fnmax = FreeCAD.ParamGet(\
        "User parameter:BaseApp/Preferences/Mod/OpenSCAD").\
        GetInt('useMaxFN', 16)
    if printverbose: print(p[3])
    if h > 0:
        if ( r1 == r2 and r1 > 0):
            if printverbose: print("Make Cylinder")
            if n < 3 or fnmax != 0 and n > fnmax:
                mycyl=doc.addObject("Part::Cylinder",p[1])
                mycyl.Height = h
                mycyl.Radius = r1
            else :
                if printverbose: print("Make Prism")
                if False: #user Draft Polygon
                    mycyl=doc.addObject("Part::Extrusion","prism")
                    mycyl.Dir = (0,0,h)
                    try :
                        import Draft
                        mycyl.Base = Draft.makePolygon(n,r1,face=True)
                    except Exception:
                        # If Draft can't import (probably due to lack of Pivy on Mac and
                        # Linux builds of FreeCAD), this is a fallback.
                        # or old level of FreeCAD
                        if printverbose:
                            print("Draft makePolygon Failed, falling back on manual polygon")
                        mycyl.Base = myPolygon(n,r1)

                    else :
                        pass
                    if gui:
                        mycyl.Base.ViewObject.hide()
                else: #Use Part::Prism primitive
                    mycyl=doc.addObject("Part::Prism","prism")
                    mycyl.Polygon = n
                    mycyl.Circumradius  = r1
                    mycyl.Height  = h

        elif (r1 != r2):
            if n < 3 or fnmax != 0 and n > fnmax:
                if printverbose: print("Make Cone")
                mycyl=doc.addObject("Part::Cone",p[1])
                mycyl.Height = h
                mycyl.Radius1 = r1
                mycyl.Radius2 = r2
            else:
                if printverbose: print("Make Frustum")
                mycyl=doc.addObject("Part::FeaturePython",'frustum')
                Frustum(mycyl,r1,r2,n,h)
                if gui:
                    if FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/OpenSCAD").\
                        GetBool('useViewProviderTree'):
                        from OpenSCADFeatures import ViewProviderTree
                        ViewProviderTree(mycyl.ViewObject)
                    else:
                        mycyl.ViewObject.Proxy = 0
        else: # r1 == r2 == 0
            FreeCAD.Console.PrintWarning('cylinder with radius zero\n')
            mycyl=doc.addObject("Part::Feature","emptycyl")
            mycyl.Shape = Part.Compound([])
    else: # h == 0
        FreeCAD.Console.PrintWarning('cylinder with height <= zero\n')
        mycyl=doc.addObject("Part::Feature","emptycyl")
        mycyl.Shape = Part.Compound([])
    if printverbose: print("Center = ",tocenter)
    if tocenter=='true' :
        center(mycyl,0,0,h)
    if False :
#   Does not fix problemfile or beltTighener although later is closer
        newobj=doc.addObject("Part::FeaturePython",'RefineCylinder')
        RefineShape(newobj,mycyl)
        if gui:
            if FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/OpenSCAD").\
                GetBool('useViewProviderTree'):
                from OpenSCADFeatures import ViewProviderTree
                ViewProviderTree(newobj.ViewObject)
            else:
                newobj.ViewObject.Proxy = 0
            mycyl.ViewObject.hide()
        p[0] = [newobj]
    else :
        p[0] = [mycyl]
    if printverbose: print("End Cylinder")

def p_cube_action(p):
    'cube_action : cube LPAREN keywordargument_list RPAREN SEMICOL'
    global doc
    l,w,h = [float(str1) for str1 in p[3]['size']]
    if (l > 0 and w > 0 and h >0):
        if printverbose: print("cube : ",p[3])
        mycube=doc.addObject('Part::Box',p[1])
        mycube.Length=l
        mycube.Width=w
        mycube.Height=h
    else:
        FreeCAD.Console.PrintWarning('cube with radius zero\n')
        mycube=doc.addObject("Part::Feature","emptycube")
        mycube.Shape = Part.Compound([])
    if p[3].get('center','false')=='true' :
        center(mycube,l,w,h)
    p[0] = [mycube]
    if printverbose: print("End Cube")

def p_circle_action(p) :
    'circle_action : circle LPAREN keywordargument_list RPAREN SEMICOL'
    if printverbose: print("Circle : "+str(p[3]))
    r = float(p[3]['r'])
    # Avoid zero radius
    if r == 0 : r = 0.00001
    n = int(p[3]['$fn'])
    fnmax = FreeCAD.ParamGet(\
        "User parameter:BaseApp/Preferences/Mod/OpenSCAD").\
        GetInt('useMaxFN',16)
    # Alter Max polygon to control if polygons are circles or polygons
    # in the modules preferences
    import Draft
    if n == 0 or fnmax != 0 and n >= fnmax:
        mycircle = FreeCAD.ActiveDocument.addObject("Part::Part2DObjectPython",'circle')
        Draft._Circle(mycircle)
        mycircle.Radius = r
        mycircle.MakeFace = True
        mycircle = Draft.makeCircle(r,face=True) # would call doc.recompute
        FreeCAD.ActiveDocument.recompute()
    else :
        mycircle = FreeCAD.ActiveDocument.addObject("Part::Part2DObjectPython",'polygon')
        Draft._Polygon(mycircle)
        mycircle.FacesNumber = n
        mycircle.Radius = r
        mycircle.DrawMode = "inscribed"
        mycircle.MakeFace = True
    if gui:
        Draft._ViewProviderDraft(mycircle.ViewObject)
    if printverbose: print("Push Circle")
    p[0] = [mycircle]

def p_square_action(p) :
    'square_action : square LPAREN keywordargument_list RPAREN SEMICOL'
    if printverbose: print("Square")
    size = p[3]['size']
    x = float(size[0])
    y = float(size[1])
    mysquare = doc.addObject('Part::Plane',p[1])
    mysquare.Length=x
    mysquare.Width=y
    if p[3].get('center','false')=='true' :
        center(mysquare,x,y,0)
    p[0] = [mysquare]

def addString(t,s,p):
    return(t + ', ' +s+' = "'+p[3][s]+'"')

def addValue(t,v,p):
    return(t + ', ' +v+' = '+p[3][v])

def p_text_action(p) :
    'text_action : text LPAREN keywordargument_list RPAREN SEMICOL'
    # If text string is null ignore
    if p[3]['text'] == "" or p[3]['text'] == " " :
        p[0] = []
        return
    t = 'text ( text="'+p[3]['text']+'"'
    t = addValue(t,'size',p)
    t = addString(t,'spacing',p)
    t = addString(t,'font',p)
    t = addString(t,'direction',p)
    t = addString(t,'language',p)
    if "script" in p[3]:
        t = addString(t,'script',p)
    else:
        t += ', script="latin"'
    t = addString(t,'halign',p)
    t = addString(t,'valign',p)
    t = addValue(t,'$fn',p)
    t = addValue(t,'$fa',p)
    t = addValue(t,'$fs',p)
    t = t+');'

    FreeCAD.Console.PrintMessage("textmsg : "+t+"\n")
    p[0] = [processTextCmd(t)]

def convert_points_list_to_vector(l):
    v = []
    for i in l :
        if printverbose: print(i)
        v.append(FreeCAD.Vector(i[0],i[1]))
    if printverbose: print(v)
    return v


def p_polygon_action_nopath(p) :
    'polygon_action_nopath : polygon LPAREN points EQ OSQUARE points_list_2d ESQUARE COMMA paths EQ undef COMMA keywordargument_list RPAREN SEMICOL'
    if printverbose: print("Polygon")
    if printverbose: print(p[6])
    v = convert_points_list_to_vector(p[6])
    mypolygon = doc.addObject('Part::Feature',p[1])
    if printverbose: print("Make Parts")
    # Close Polygon
    v.append(v[0])
    parts = Part.makePolygon(v)
    if printverbose: print("update object")
    mypolygon.Shape = Part.Face(parts)
    p[0] = [mypolygon]

def p_polygon_action_plus_path(p) :
    'polygon_action_plus_path : polygon LPAREN points EQ OSQUARE points_list_2d ESQUARE COMMA paths EQ OSQUARE path_set ESQUARE COMMA keywordargument_list RPAREN SEMICOL'
    if printverbose: print("Polygon with Path")
    if printverbose: print(p[6])
    v = convert_points_list_to_vector(p[6])
    if printverbose: print("Path Set List")
    if printverbose: print(p[12])
    for i in p[12] :
        if printverbose: print(i)
        mypolygon = doc.addObject('Part::Feature','wire')
        path_list = []
        for j in i :
            j = int(j)
            if printverbose: print(j)
            path_list.append(v[j])
#       Close path
        path_list.append(v[int(i[0])])
        if printverbose: print('Path List')
        if printverbose: print(path_list)
        wire = Part.makePolygon(path_list)
        mypolygon.Shape = Part.Face(wire)
        p[0] = [mypolygon]
#       This only pushes last polygon

def make_face(v1,v2,v3):
    wire = Part.makePolygon([v1,v2,v3,v1])
    face = Part.Face(wire)
    return face

def p_polyhedron_action(p) :
    '''polyhedron_action : polyhedron LPAREN points EQ OSQUARE points_list_3d ESQUARE COMMA faces EQ OSQUARE path_set ESQUARE COMMA keywordargument_list RPAREN SEMICOL
                      | polyhedron LPAREN points EQ OSQUARE points_list_3d ESQUARE COMMA triangles EQ OSQUARE points_list_3d ESQUARE COMMA keywordargument_list RPAREN SEMICOL'''
    if printverbose: print("Polyhedron Points")
    v = []
    for i in p[6] :
        if printverbose: print(i)
        v.append(FreeCAD.Vector(float(i[0]),float(i[1]),float(i[2])))
    if printverbose:
        print(v)
        print ("Polyhedron "+p[9])
        print (p[12])
    faces_list = []
    mypolyhed = doc.addObject('Part::Feature',p[1])
    for i in p[12] :
        if printverbose: print(i)
        v2 = FreeCAD.Vector
        pp =[v2(v[k]) for k in i]
        # Add first point to end of list to close polygon
        pp.append(pp[0])
        try:
            w = Part.makePolygon(pp)
            f = Part.Face(w)
        except Exception:
            secWireList = w.Edges[:]
            f = Part.makeFilledFace(Part.__sortEdges__(secWireList))
        faces_list.append(f)
    shell=Part.makeShell(faces_list)
    solid=Part.Solid(shell).removeSplitter()
    if solid.Volume < 0:
        solid.reverse()
    mypolyhed.Shape=solid
    p[0] = [mypolyhed]

def p_projection_action(p) :
    'projection_action : projection LPAREN keywordargument_list RPAREN OBRACE block_list EBRACE'
    if printverbose: print('Projection')

    doc.recompute()
    p[6][0].Shape.tessellate(0.05) # Ensure the bounding box calculation is not done with the splines, which can give a bad result
    bbox = p[6][0].Shape.BoundBox
    for shape in p[6]:
        shape.Shape.tessellate(0.05)
        bbox.add(shape.Shape.BoundBox)
    plane = doc.addObject("Part::Plane","xy_plane_used_for_projection")
    plane.Length = bbox.XLength
    plane.Width = bbox.YLength
    plane.Placement = FreeCAD.Placement(FreeCAD.Vector(\
                     bbox.XMin,bbox.YMin,0),FreeCAD.Rotation())
    if gui:
        plane.ViewObject.hide()

    if p[3]['cut'] == 'true' :
        obj = doc.addObject('Part::MultiCommon','projection_cut')
        if (len(p[6]) > 1):
            subobj = [fuse(p[6],"projection_cut_implicit_group")]
        else:
            subobj = p[6]
        obj.Shapes = [plane] + subobj
        if gui:
            subobj[0].ViewObject.hide()
        p[0] = [obj]
    else: # cut == 'false' => true projection
        if gui and not FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/OpenSCAD").\
                GetBool('usePlaceholderForUnsupported'):
            from PySide import QtGui
            QtGui.QMessageBox.critical(None, translate('OpenSCAD',"Unsupported Function") + " : " + p[1],translate('OpenSCAD',"Press OK"))
        else:
            p[0] = [placeholder(p[1],p[6],p[3])]



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2025 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

"""Classes and utility functions to generate a remotely hosted cache of all addon catalog entries.
Intended to be run by a server-side systemd timer to generate a file that is then loaded by the
Addon Manager in each FreeCAD installation."""
import datetime
import shutil
import sys
from dataclasses import is_dataclass, fields
from typing import Any, List, Optional

import base64
import enum
import hashlib
import io
import json
import os
import requests
import subprocess
from typing import List
import xml.etree.ElementTree
import zipfile

import AddonCatalog
import addonmanager_metadata
import addonmanager_utilities as utils


ADDON_CATALOG_URL = "https://raw.githubusercontent.com/FreeCAD/Addons/main/Data/Index.json"
BASE_DIRECTORY = "./CatalogCache"
MAX_COUNT = 10000  # Do at most this many repos (for testing purposes this can be made smaller)
CLONE_TIMEOUT = (
    300  # Seconds: repos that take longer than this are assumed to be too large to index
)

# Repos that are too large, or that should for some reason not be fully cloned here
FORCE_SPARSE_CLONE = ["parts_library", "offline-documentation", "FreeCAD-Documentation-html"]


def recursive_serialize(obj: Any):
    """Recursively serialize an object, supporting non-dataclasses that themselves contain
    dataclasses (in this case, AddonCatalog, which contains AddonCatalogEntry)"""
    if is_dataclass(obj):
        result = {}
        for f in fields(obj):
            value = getattr(obj, f.name)
            result[f.name] = recursive_serialize(value)
        return result
    elif isinstance(obj, list):
        return [recursive_serialize(i) for i in obj]
    elif isinstance(obj, dict):
        return {k: recursive_serialize(v) for k, v in obj.items()}
    elif hasattr(obj, "__dict__"):
        return {k: recursive_serialize(v) for k, v in vars(obj).items() if not k.startswith("__")}
    else:
        return obj


class GitRefType(enum.IntEnum):
    """Enum for the type of git ref (tag, branch, or hash)."""

    TAG = 1
    BRANCH = 2
    HASH = 3


class CatalogFetcher:
    """Fetches the addon index from the given URL and returns an AddonCatalog object. Separated
    from the main class for easy mocking during tests. Note that every instantiation of this class
    will run a new fetch of the catalog."""

    def __init__(self, addon_catalog_url: str = ADDON_CATALOG_URL):
        self.addon_catalog_url = addon_catalog_url
        self.catalog = self.fetch_catalog()

    def fetch_catalog(self) -> AddonCatalog.AddonCatalog:
        """Fetch the addon catalog from the given URL and return an AddonCatalog object."""
        response = requests.get(self.addon_catalog_url, timeout=10.0)
        if response.status_code != 200:
            raise RuntimeError(f"ERROR: Failed to fetch addon index from {self.addon_catalog_url}")
        return AddonCatalog.AddonCatalog(response.json())


class CacheWriter:
    """Writes a JSON file containing a cache of all addon catalog entries. The cache is a copy of
    the package.xml, requirements.txt, and metadata.txt files from the addon repositories, as well
    as a base64-encoded icon image. The cache is written to the current working directory."""

    def __init__(self):
        self.catalog: Optional[AddonCatalog.AddonCatalog] = None
        self.icon_errors = {}
        self.clone_errors = {}
        if os.path.isabs(BASE_DIRECTORY):
            self.cwd = BASE_DIRECTORY
        else:
            self.cwd = os.path.normpath(os.path.join(os.getcwd(), BASE_DIRECTORY))
        self._cache = {}

    def write(self, addon_id: Optional[str] = None) -> None:
        original_working_directory = os.getcwd()
        os.makedirs(self.cwd, exist_ok=True)
        os.chdir(self.cwd)

        try:
            fetcher = CatalogFetcher()
            self.catalog = fetcher.catalog

            if addon_id is None:
                self.create_local_copy_of_addons()
            else:
                catalog = self.catalog.get_catalog()
                if addon_id not in catalog:
                    raise RuntimeError(f"ERROR: Addon {addon_id} not in index")
                catalog_entries = catalog[addon_id]
                self.create_local_copy_of_single_addon(addon_id, catalog_entries)

            # Write the entire index for versions of the Addon Manager after 2026-01-24
            with zipfile.ZipFile(
                os.path.join(self.cwd, "addon_index_cache.zip"), "w", zipfile.ZIP_DEFLATED
            ) as zipf:
                zipf.writestr(
                    "addon_index_cache.json",
                    json.dumps(recursive_serialize(self.catalog.get_catalog()), indent="  "),
                )

            # Also generate the sha256 hash of the zip file and store it
            with open("addon_index_cache.zip", "rb") as cache_file:
                cache_file_content = cache_file.read()
            sha256 = hashlib.sha256(cache_file_content).hexdigest()
            with open("addon_index_cache.zip.sha256", "w", encoding="utf-8") as hash_file:
                hash_file.write(sha256)

            # For pre-2026-01-24 write only curated addons into a separate catalog file so older
            # versions of the Addon Manager don't accidentally install uncurated addons.
            with zipfile.ZipFile(
                os.path.join(self.cwd, "addon_catalog_cache.zip"), "w", zipfile.ZIP_DEFLATED
            ) as zipf:
                catalog = self.catalog.get_catalog()
                reduced_catalog = {}
                for addon_id, catalog_entries in catalog.items():
                    approved_entries: List[AddonCatalog.AddonCatalogEntry] = []
                    for entry in catalog_entries:
                        if entry.curated:
                            approved_entries.append(entry)
                    if approved_entries:
                        reduced_catalog[addon_id] = approved_entries
                zipf.writestr(
                    "addon_catalog_cache.json",
                    json.dumps(recursive_serialize(reduced_catalog), indent="  "),
                )

            # Also generate the sha256 hash of the zip file and store it
            with open("addon_catalog_cache.zip", "rb") as cache_file:
                cache_file_content = cache_file.read()
            sha256 = hashlib.sha256(cache_file_content).hexdigest()
            with open("addon_catalog_cache.zip.sha256", "w", encoding="utf-8") as hash_file:
                hash_file.write(sha256)

            with open(os.path.join(self.cwd, "icon_errors.json"), "w") as f:
                json.dump(self.icon_errors, f, indent="  ")

            with open(os.path.join(self.cwd, "clone_errors.json"), "w") as f:
                json.dump(self.clone_errors, f, indent="  ")

            print(f"Wrote index to {os.path.join(self.cwd, 'addon_index_cache.zip')}")
            print(f"Wrote cache to {os.path.join(self.cwd, 'addon_catalog_cache.zip')}")
        finally:
            os.chdir(original_working_directory)

    def create_local_copy_of_addons(self):
        counter = 0
        for addon_id, catalog_entries in self.catalog.get_catalog().items():
            self.create_local_copy_of_single_addon(addon_id, catalog_entries)
            counter += 1
            if counter >= MAX_COUNT:
                break

    def create_local_copy_of_single_addon(
        self, addon_id: str, catalog_entries: List[AddonCatalog.AddonCatalogEntry]
    ):
        for index, catalog_entry in enumerate(catalog_entries):
            if addon_id in FORCE_SPARSE_CLONE:
                if catalog_entry.repository is None:
                    print(
                        f"ERROR: Cannot use sparse clone for {addon_id} because it has no git repo."
                    )
                    continue
                if catalog_entry.zip_url is None:
                    print(
                        f"ERROR: Cannot use sparse clone for {addon_id} because it has no zip URL."
                    )
                    continue
                catalog_entry.sparse_cache = True
                self.create_local_copy_of_single_addon_with_git_sparse(
                    addon_id, index, catalog_entry
                )
            elif catalog_entry.repository is not None:
                self.create_local_copy_of_single_addon_with_git(addon_id, index, catalog_entry)
            elif catalog_entry.zip_url is not None:
                self.create_local_copy_of_single_addon_with_zip(addon_id, index, catalog_entry)
            else:
                print(
                    f"ERROR: Invalid catalog entry for {addon_id}. "
                    "Neither git info nor zip info was specified."
                )
                continue
            metadata = self.generate_cache_entry(addon_id, index, catalog_entry)
            self.catalog.add_metadata_to_entry(addon_id, index, metadata)
            self.create_zip_of_entry(addon_id, index, catalog_entry)

    def generate_cache_entry(
        self, addon_id: str, index: int, catalog_entry: AddonCatalog.AddonCatalogEntry
    ) -> Optional[AddonCatalog.CatalogEntryMetadata]:
        """Create the cache entry for this catalog entry if there is data to cache. If there is
        nothing to cache, returns None."""
        path_to_package_xml = self.find_file("package.xml", addon_id, index, catalog_entry)
        cache_entry = None
        if path_to_package_xml and os.path.exists(path_to_package_xml):
            cache_entry = self.generate_cache_entry_from_package_xml(path_to_package_xml)

        path_to_requirements = self.find_file("requirements.txt", addon_id, index, catalog_entry)
        if path_to_requirements and os.path.exists(path_to_requirements):
            if cache_entry is None:
                cache_entry = AddonCatalog.CatalogEntryMetadata()
            with open(path_to_requirements, "r", encoding="utf-8") as f:
                cache_entry.requirements_txt = f.read()

        path_to_metadata = self.find_file("metadata.txt", addon_id, index, catalog_entry)
        if path_to_metadata and os.path.exists(path_to_metadata):
            if cache_entry is None:
                cache_entry = AddonCatalog.CatalogEntryMetadata()
            with open(path_to_metadata, "r", encoding="utf-8") as f:
                cache_entry.metadata_txt = f.read()

        dirname = CacheWriter.get_directory_name(addon_id, index, catalog_entry)
        if os.path.exists(os.path.join(self.cwd, dirname, ".git")):
            old_dir = os.getcwd()
            os.chdir(os.path.join(self.cwd, dirname))
            last_updated_time = CacheWriter.determine_last_commit_time()
            if last_updated_time:
                catalog_entry.last_update_time = last_updated_time.isoformat()
            os.chdir(old_dir)

        zip_name = os.path.join(self.cwd, dirname + ".zip")
        if os.path.exists(zip_name):
            # Don't use os.path.join, by convention this path is always UNIX style, and local
            # users are required to translate it into their OS's format as needed
            catalog_entry.relative_cache_path = BASE_DIRECTORY + "/" + dirname + ".zip"

        return cache_entry

    def generate_cache_entry_from_package_xml(
        self, path_to_package_xml: str
    ) -> Optional[AddonCatalog.CatalogEntryMetadata]:
        cache_entry = AddonCatalog.CatalogEntryMetadata()
        with open(path_to_package_xml, "r", encoding="utf-8") as f:
            cache_entry.package_xml = f.read()
        try:
            metadata = addonmanager_metadata.MetadataReader.from_bytes(
                cache_entry.package_xml.encode("utf-8")
            )
        except xml.etree.ElementTree.ParseError:
            print(f"ERROR: Failed to parse XML from {path_to_package_xml}")
            return None
        except RuntimeError:
            print(f"ERROR: Failed to read metadata from {path_to_package_xml}")
            return None

        relative_icon_path = self.get_icon_from_metadata(metadata)
        if relative_icon_path is not None:
            absolute_icon_path = os.path.join(
                os.path.dirname(path_to_package_xml), relative_icon_path
            )
            if os.path.exists(absolute_icon_path):
                with open(absolute_icon_path, "rb") as f:
                    try:
                        cache_entry.icon_data = base64.b64encode(f.read()).decode("utf-8")
                    except IOError as e:
                        print(f"ERROR: IO Error while reading icon file {absolute_icon_path}")
                        print(e)
                    except Exception as e:
                        print(f"ERROR: Unknown error while reading icon file {absolute_icon_path}")
                        print(e)
            else:
                self.icon_errors[metadata.name] = relative_icon_path
                print(f"ERROR: Could not find icon file {absolute_icon_path}")
        return cache_entry

    def create_local_copy_of_single_addon_with_git(
        self, addon_id: str, index: int, catalog_entry: AddonCatalog.AddonCatalogEntry
    ):
        expected_name = self.get_directory_name(addon_id, index, catalog_entry)
        try:
            self.clone_or_update(expected_name, catalog_entry.repository, catalog_entry.git_ref)
        except RuntimeError as e:
            print(f"ERROR: Failed to clone or update {addon_id} from {catalog_entry.repository}.")
            print(f"ERROR: {e}")

    def create_local_copy_of_single_addon_with_git_sparse(
        self, addon_id: str, index: int, catalog_entry: AddonCatalog.AddonCatalogEntry
    ):
        expected_name = self.get_directory_name(addon_id, index, catalog_entry)
        try:
            files = ["package.xml", "requirements.txt", "metadata.txt"]
            self.sparse_clone(expected_name, catalog_entry.repository, catalog_entry.git_ref, files)
            if os.path.exists(os.path.join(self.cwd, expected_name, "package.xml")):
                metadata = addonmanager_metadata.MetadataReader.from_file(
                    os.path.join(self.cwd, expected_name, "package.xml")
                )
                if metadata.icon:
                    self.add_to_sparse_clone(expected_name, [metadata.icon])
        except RuntimeError as e:
            print(f"ERROR: Failed to clone or update {addon_id} from {catalog_entry.repository}.")
            print(f"ERROR: {e}")

    @staticmethod
    def get_directory_name(addon_id, index, catalog_entry):
        expected_name = os.path.join(addon_id, str(index) + "-")
        if catalog_entry.branch_display_name:
            expected_name += catalog_entry.branch_display_name.replace("/", "-")
        elif catalog_entry.git_ref:
            expected_name += catalog_entry.git_ref.replace("/", "-")
        else:
            expected_name += "unknown-branch-name"
        return expected_name

    def create_local_copy_of_single_addon_with_zip(
        self, addon_id: str, index: int, catalog_entry: AddonCatalog.AddonCatalogEntry
    ):
        response = requests.get(catalog_entry.zip_url, timeout=10.0)
        if response.status_code != 200:
            print(f"ERROR: Failed to fetch zip data for {addon_id} from {catalog_entry.zip_url}.")
            return
        extract_to_dir = self.get_directory_name(addon_id, index, catalog_entry)
        if os.path.exists(extract_to_dir):
            utils.rmdir(extract_to_dir)
        os.makedirs(extract_to_dir, exist_ok=True)

        with zipfile.ZipFile(io.BytesIO(response.content)) as zip_file:
            latest = max(
                (info.date_time for info in zip_file.infolist() if not info.is_dir()), default=None
            )
            if latest is not None:
                catalog_entry.last_update_time = datetime.datetime(*latest).isoformat()
            zip_file.extractall(path=extract_to_dir)

    def clone_or_update(self, name: str, url: str, branch: str) -> None:
        """If a directory called "name" exists, and it contains a subdirectory called .git,
        then 'git fetch' is called; otherwise we use 'git clone' to make a bare, shallow
        copy of the repo (in the normal case where minimal is True), or a normal clone,
        if minimal is set to False."""

        if not os.path.exists(os.path.join(os.getcwd(), name, ".git")):
            print(f"Cloning {url} to {name}", flush=True)
            # Shallow, but do include the last commit on each branch and tag
            command = [
                "git",
                "clone",
                "--depth",
                "1",
                "--branch",
                branch,
                url,
                name,
            ]
            try:
                completed_process = subprocess.run(command, timeout=CLONE_TIMEOUT)
            except subprocess.TimeoutExpired:
                self.clone_errors[name] = f"Timed out after {CLONE_TIMEOUT} seconds."
                raise RuntimeError(f"Clone of {url} timed out.")
                # TODO: Automatically fall back to a sparse clone
            if completed_process.returncode != 0:
                self.clone_errors[name] = f"Failed to clone {url}: {completed_process.returncode}"
                raise RuntimeError(f"Clone failed for {url}")
        else:
            print(f"Updating {name}", flush=True)
            old_dir = os.getcwd()
            os.chdir(os.path.join(old_dir, name))
            try:
                # Determine if we are dealing with a tag, branch, or hash
                git_ref_type = CacheWriter.determine_git_ref_type(name, url, branch)
                command = ["git", "fetch"]
                completed_process = subprocess.run(command)
                if completed_process.returncode != 0:
                    os.chdir(old_dir)
                    raise RuntimeError(f"git fetch failed for {name}")
                command = ["git", "checkout", branch, "--quiet"]
                completed_process = subprocess.run(command)
                if completed_process.returncode != 0:
                    os.chdir(old_dir)
                    raise RuntimeError(f"git checkout failed for {name} branch {branch}")
                if git_ref_type == GitRefType.BRANCH:
                    command = ["git", "merge", "--quiet"]
                    completed_process = subprocess.run(command)
                    if completed_process.returncode != 0:
                        os.chdir(old_dir)
                        raise RuntimeError(f"git merge failed for {name} branch {branch}")
            except RuntimeError as e:
                # In the event of basically ANY error, delete the original and re-clone.
                print(e)
                print("Deleting and re-cloning the original repo")
                os.chdir(old_dir)
                utils.rmdir(os.path.join(old_dir, name))
                self.clone_or_update(name, url, branch)
            os.chdir(old_dir)

    def sparse_clone(self, name: str, url: str, branch: str, files: List[str]) -> None:
        """Perform a sparse clone of a git repo, including only the specified files. Overwrite any
        existing path."""

        if not os.path.exists(os.path.join(os.getcwd(), name, ".git")):
            print(f"Creating sparse clone {name}", flush=True)
            cwd = os.getcwd()
            clone_path = os.path.join(cwd, name)
            if os.path.exists(clone_path):
                try:
                    shutil.rmtree(clone_path)
                except OSError as e:
                    self.clone_errors[name] = f"Failed to remove existing path {clone_path}: {e}"
                    print(f"ERROR: Failed to remove existing path {clone_path}: {e}")
                    return
            os.makedirs(clone_path)
            os.chdir(clone_path)
            try:
                subprocess.run(["git", "init", "--quiet"], check=True)
                subprocess.run(["git", "remote", "add", "origin", url], check=True)
                subprocess.run(["git", "config", "core.sparsecheckout", "true"], check=True)
                with open(".git/info/sparse-checkout", "w") as f:
                    f.write("\n".join(files))
                    f.write("\n")  # So we are safe appending later
                subprocess.run(
                    ["git", "fetch", "--depth=1", "origin", branch],
                    check=True,
                    timeout=CLONE_TIMEOUT,
                )
                subprocess.run(["git", "checkout", branch], check=True)
            except (subprocess.CalledProcessError, subprocess.TimeoutExpired) as e:
                self.clone_errors[name] = str(e)
                print(f"ERROR: {e}")
            os.chdir(cwd)
        else:
            print(f"Updating sparse clone {name}", flush=True)
            cwd = os.getcwd()
            os.chdir(os.path.join(cwd, name))
            try:
                subprocess.run(["git", "pull", "--depth=1"], check=True)
            except (subprocess.CalledProcessError, subprocess.TimeoutExpired) as e:
                self.clone_errors[name] = str(e)
                print(f"ERROR: {e}")
            os.chdir(cwd)

    def add_to_sparse_clone(self, name: str, files: List[str]) -> None:
        """Clones additional files to an existing sparse clone."""
        cwd = os.getcwd()
        clone_path = os.path.join(cwd, name)
        os.chdir(clone_path)
        with open(".git/info/sparse-checkout", "a") as f:
            f.write("\n".join(files))
        try:
            subprocess.run(["git", "pull", "--depth=1"], check=True)
        except subprocess.CalledProcessError as e:
            self.clone_errors[name] = str(e)
            print(f"ERROR: {e}")
        os.chdir(cwd)

    def find_file(
        self,
        filename: str,
        addon_id: str,
        index: int,
        catalog_entry: AddonCatalog.AddonCatalogEntry,
    ) -> Optional[str]:
        """Find a given file in the downloaded cache for this addon. Returns None if the file does
        not exist."""
        start_dir = os.path.join(self.cwd, self.get_directory_name(addon_id, index, catalog_entry))
        for dirpath, _, filenames in os.walk(start_dir):
            if filename in filenames:
                return os.path.join(dirpath, filename)
        return None

    @staticmethod
    def get_icon_from_metadata(metadata: addonmanager_metadata.Metadata) -> Optional[str]:
        """Try to locate the icon file specified for this Addon. Recursively search through the
        levels of the metadata and return the first specified icon file path. Returns None of there
        is no icon specified for this Addon (which is not allowed by the standard, but we don't want
        to crash the cache writer)."""
        if metadata.icon:
            if (
                metadata.subdirectory
                and metadata.subdirectory != "."
                and metadata.subdirectory != "./"
            ):
                return metadata.subdirectory + "/" + metadata.icon
            return metadata.icon
        for content_type in metadata.content:
            for content_item in metadata.content[content_type]:
                icon = CacheWriter.get_icon_from_metadata(content_item)
                if icon:
                    return icon
        return None

    @staticmethod
    def determine_git_ref_type(name: str, _url: str, branch: str) -> GitRefType:
        """Determine if the given branch, tag, or hash is a tag, branch, or hash. Returns the type
        if determinable, otherwise raises a RuntimeError."""
        command = ["git", "show-ref", "--verify", f"refs/remotes/origin/{branch}"]
        completed_process = subprocess.run(command, capture_output=True)
        if completed_process.returncode == 0:
            return GitRefType.BRANCH
        command = ["git", "show-ref", "--tags"]
        completed_process = subprocess.run(command, capture_output=True)
        completed_process_output = completed_process.stdout.decode("utf-8")
        if branch in completed_process_output:
            return GitRefType.TAG
        command = ["git", "rev-parse", branch]
        completed_process = subprocess.run(command)
        if completed_process.returncode == 0:
            return GitRefType.HASH
        raise RuntimeError(
            f"Could not determine if {branch} of {name} is a tag, branch, or hash. "
            f"Output was: {completed_process_output}"
        )

    @staticmethod
    def determine_last_commit_time() -> datetime.datetime:
        """Executed on the current working directory. Returns the time of the last commit."""
        command = ["git", "log", "-1", "--format=%cd", "--date=iso-strict"]
        completed_process = subprocess.run(command, capture_output=True)
        completed_process_output = completed_process.stdout.decode("utf-8").strip()
        try:
            dt = datetime.datetime.fromisoformat(completed_process_output)
        except ValueError:
            print(f"ERROR: Failed to parse last commit time from {completed_process_output}")
            dt = None
        return dt

    def create_zip_of_entry(
        self, addon_id: str, index: int, catalog_entry: AddonCatalog.AddonCatalogEntry
    ):
        """Create a zip file containing the contents of the addon directory for this entry. The
        zip file is written to a file with the same name as the calculated addon cache directory
        in the current working directory."""

        dirname = CacheWriter.get_directory_name(addon_id, index, catalog_entry)
        start_dir = os.path.join(self.cwd, dirname)
        zip_file_path = os.path.join(self.cwd, f"{dirname}.zip")
        temp_file_path = zip_file_path + ".new"

        if not os.path.isdir(start_dir):
            print(
                f"ERROR: Directory {start_dir} does not exist. Skipping zip creation for addon {addon_id}."
            )
            return

        with zipfile.ZipFile(temp_file_path, "w", zipfile.ZIP_DEFLATED) as zf:
            for root, dirs, files in os.walk(start_dir):
                if ".git" in dirs:
                    dirs.remove(".git")  # Don't visit .git directories
                for file in files:
                    full_path = os.path.join(root, file)
                    rel_path = os.path.relpath(full_path, start_dir)
                    try:
                        zf.write(full_path, rel_path)
                    except (OSError, FileNotFoundError, RuntimeError) as e:
                        print(f"WARNING: Could not add {full_path} to zip archive: {e}")
        try:
            good = False
            with zipfile.ZipFile(temp_file_path, "r") as zf:
                good = zf.testzip() is None
            if good:
                if os.path.exists(zip_file_path):
                    os.remove(zip_file_path)
                os.rename(temp_file_path, zip_file_path)
            else:
                os.remove(temp_file_path)
                print(
                    f"Failed to create zip file {zip_file_path} for addon {addon_id}: data is corrupt"
                )
        except zipfile.BadZipFile:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)
            print(
                f"Failed to create zip file {zip_file_path} for addon {addon_id}: data is not a valid zip file"
            )


if __name__ == "__main__":
    single_addon_id = None
    if len(sys.argv) > 1:
        single_addon_id = sys.argv[1]
    writer = CacheWriter()
    writer.write(single_addon_id)



# SPDX-License-Identifier: LGPL-2.1-or-later

from .browser import LibraryBrowserWidget
from .dock import ToolBitLibraryDock
from .editor import LibraryEditor
from .properties import LibraryPropertyDialog


__all__ = [
    "LibraryBrowserWidget",
    "ToolBitLibraryDock",
    "LibraryEditor",
    "LibraryPropertyDialog",
]



# ***************************************************************************
# *   Copyright (c) 2014 Juergen Riegel <FreeCAD@juergen-riegel.net>        *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD, FreeCADGui, Sketcher, Part


__title__ = "Hexagon profile lib"
__author__ = "Juergen Riegel"
__url__ = "https://www.freecad.org"

App = FreeCAD
Gui = FreeCADGui


def makeHexagonSimple(sketchName=None):
    if not sketchName:
        sketch = App.ActiveDocument.addObject("Sketcher::SketchObject", "Hexagon")
    else:
        sketch = App.ActiveDocument.getObject(sketchName)

    geoList = []
    geoList.append(Part.LineSegment(App.Vector(-20.00, 34.64, 0), App.Vector(20.00, 34.64, 0)))
    geoList.append(Part.LineSegment(App.Vector(20.00, 34.64, 0), App.Vector(47.082363, 0.00, 0)))
    geoList.append(Part.LineSegment(App.Vector(40.00, 0.00, 0), App.Vector(20.00, -34.64, 0)))
    geoList.append(Part.LineSegment(App.Vector(20.00, -34.64, 0), App.Vector(-20.00, -34.64, 0)))
    geoList.append(Part.LineSegment(App.Vector(-20.00, -34.64, 0), App.Vector(-40.00, 0.00, 0)))
    geoList.append(Part.LineSegment(App.Vector(-40.00, 0.00, 0), App.Vector(-20.00, 34.64, 0)))
    (l1, l2, l3, l4, l5, l6) = sketch.addGeometry(geoList)

    conList = []
    conList.append(Sketcher.Constraint("Coincident", l1, 2, l2, 1))
    conList.append(Sketcher.Constraint("Coincident", l2, 2, l3, 1))
    conList.append(Sketcher.Constraint("Coincident", l3, 2, l4, 1))
    conList.append(Sketcher.Constraint("Coincident", l4, 2, l5, 1))
    conList.append(Sketcher.Constraint("Coincident", l5, 2, l6, 1))
    conList.append(Sketcher.Constraint("Coincident", l6, 2, l1, 1))
    conList.append(Sketcher.Constraint("Equal", l1, l2))
    conList.append(Sketcher.Constraint("Equal", l2, l3))
    conList.append(Sketcher.Constraint("Equal", l3, l4))
    conList.append(Sketcher.Constraint("Equal", l4, l5))
    conList.append(Sketcher.Constraint("Equal", l5, l6))
    conList.append(Sketcher.Constraint("Angle", l1, 2, l2, 1, App.Units.Quantity("120.000000 deg")))
    conList.append(Sketcher.Constraint("Angle", l3, 2, l4, 1, App.Units.Quantity("120.000000 deg")))
    conList.append(Sketcher.Constraint("Angle", l5, 2, l6, 1, App.Units.Quantity("120.000000 deg")))
    sketch.addConstraint(conList)
    return



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2016 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD SweetHome3D Importer"
__author__ = "Yorik van Havre"
__url__ = "https://www.freecad.org"

## @package importSH3D
#  \ingroup ARCH
#  \brief SH3D (SweetHome3D) file format importer
#
#  This module provides tools to import SH3D files created from Sweet Home 3D.

import os
import zipfile

import FreeCAD
from FreeCAD import Base


DEBUG = True


def open(filename):
    "called when freecad wants to open a file"
    docname = os.path.splitext(os.path.basename(filename))[0]
    doc = FreeCAD.newDocument(docname)
    doc.Label = docname
    FreeCAD.ActiveDocument = doc
    read(filename)
    return doc


def insert(filename, docname):
    "called when freecad wants to import a file"
    try:
        doc = FreeCAD.getDocument(docname)
    except NameError:
        doc = FreeCAD.newDocument(docname)
    FreeCAD.ActiveDocument = doc
    read(filename)
    return doc


def read(filename):
    "reads the file and creates objects in the active document"

    import BIM.importers.importSH3DHelper

    if DEBUG:
        from importlib import reload

        reload(BIM.importers.importSH3DHelper)

    pi = Base.ProgressIndicator()
    try:
        importer = BIM.importers.importSH3DHelper.SH3DImporter(pi)
        importer.import_sh3d_from_filename(filename)
    finally:
        pi.stop()

    FreeCAD.ActiveDocument.recompute()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Initialization of the Draft workbench (graphical interface)."""

import os

import FreeCAD
import FreeCADGui

__title__ = "FreeCAD Draft Workbench - Init file"
__author__ = "Yorik van Havre <yorik@uncreated.net>"
__url__ = "https://www.freecad.org"


class DraftWorkbench(FreeCADGui.Workbench):
    """The Draft Workbench definition."""

    def __init__(self):
        def QT_TRANSLATE_NOOP(context, text):
            return text

        __dirname__ = os.path.join(FreeCAD.getResourceDir(), "Mod", "Draft")
        _tooltip = "The Draft workbench is used for 2D drafting on a grid"
        self.__class__.Icon = os.path.join(__dirname__, "Resources", "icons", "DraftWorkbench.svg")
        self.__class__.MenuText = QT_TRANSLATE_NOOP("draft", "Draft")
        self.__class__.ToolTip = QT_TRANSLATE_NOOP("draft", _tooltip)

    def Initialize(self):
        """When the workbench is first loaded."""

        def QT_TRANSLATE_NOOP(context, text):
            return text

        # Run self-tests
        dependencies_OK = False
        try:
            from pivy import coin

            if FreeCADGui.getSoDBVersion() != coin.SoDB.getVersion():
                raise AssertionError(
                    "FreeCAD and Pivy use different versions "
                    "of Coin. "
                    "This will lead to unexpected behaviour."
                )
        except AssertionError:
            FreeCAD.Console.PrintWarning(
                "Error: FreeCAD and Pivy "
                "use different versions of Coin. "
                "This will lead to unexpected "
                "behaviour.\n"
            )
        except ImportError:
            FreeCAD.Console.PrintWarning(
                "Error: Pivy not found, " "Draft Workbench will be disabled.\n"
            )
        except Exception:
            FreeCAD.Console.PrintWarning("Error: Unknown error " "while trying to load Pivy.\n")
        else:
            dependencies_OK = True

        if not dependencies_OK:
            return

        # Import Draft tools, icons
        try:
            import Draft_rc
            import DraftTools
            import DraftGui

            FreeCADGui.addLanguagePath(":/translations")
            FreeCADGui.addIconPath(":/icons")
        except Exception as exc:
            FreeCAD.Console.PrintError(exc)
            FreeCAD.Console.PrintError(
                "Error: Initializing one or more "
                "of the Draft modules failed, "
                "Draft will not work as expected.\n"
            )

        # Set up command lists
        import draftutils.init_tools as it

        self.drawing_commands = it.get_draft_drawing_commands()
        self.annotation_commands = it.get_draft_annotation_commands()
        self.modification_commands = it.get_draft_modification_commands()
        self.utility_commands_menu = it.get_draft_utility_commands_menu()
        self.utility_commands_toolbar = it.get_draft_utility_commands_toolbar()
        self.context_commands = it.get_draft_context_commands()

        # Set up toolbars
        it.init_toolbar(
            self, QT_TRANSLATE_NOOP("Workbench", "Draft Creation"), self.drawing_commands
        )
        it.init_toolbar(
            self, QT_TRANSLATE_NOOP("Workbench", "Draft Annotation"), self.annotation_commands
        )
        it.init_toolbar(
            self, QT_TRANSLATE_NOOP("Workbench", "Draft Modification"), self.modification_commands
        )
        it.init_toolbar(
            self, QT_TRANSLATE_NOOP("Workbench", "Draft Utility"), self.utility_commands_toolbar
        )
        it.init_toolbar(
            self, QT_TRANSLATE_NOOP("Workbench", "Draft Snap"), it.get_draft_snap_commands()
        )

        # Set up menus
        it.init_menu(self, [QT_TRANSLATE_NOOP("Workbench", "&Drafting")], self.drawing_commands)
        it.init_menu(
            self, [QT_TRANSLATE_NOOP("Workbench", "&Annotation")], self.annotation_commands
        )
        it.init_menu(
            self, [QT_TRANSLATE_NOOP("Workbench", "&Modification")], self.modification_commands
        )
        it.init_menu(
            self, [QT_TRANSLATE_NOOP("Workbench", "&Utilities")], self.utility_commands_menu
        )

        # Set up preferences pages
        if hasattr(FreeCADGui, "draftToolBar"):
            if not hasattr(FreeCADGui.draftToolBar, "loadedPreferences"):
                from draftutils import params

                params._param_observer_start()
                FreeCADGui.addPreferencePage(
                    ":/ui/preferences-draft.ui", QT_TRANSLATE_NOOP("QObject", "Draft")
                )
                FreeCADGui.addPreferencePage(
                    ":/ui/preferences-draftinterface.ui", QT_TRANSLATE_NOOP("QObject", "Draft")
                )
                FreeCADGui.addPreferencePage(
                    ":/ui/preferences-draftsnap.ui", QT_TRANSLATE_NOOP("QObject", "Draft")
                )
                FreeCADGui.addPreferencePage(
                    ":/ui/preferences-draftvisual.ui", QT_TRANSLATE_NOOP("QObject", "Draft")
                )
                FreeCADGui.addPreferencePage(
                    ":/ui/preferences-drafttexts.ui", QT_TRANSLATE_NOOP("QObject", "Draft")
                )
                FreeCADGui.draftToolBar.loadedPreferences = True

        FreeCADGui.getMainWindow().mainWindowClosed.connect(self.Deactivated)

        FreeCAD.Console.PrintLog("Loading Draft workbench, done.\n")

    def Activated(self):
        """When entering the workbench."""
        if hasattr(FreeCADGui, "draftToolBar"):
            FreeCADGui.draftToolBar.Activated()
        if hasattr(FreeCADGui, "Snapper"):
            FreeCADGui.Snapper.show()
            from draftutils import init_draft_statusbar

            init_draft_statusbar.show_draft_statusbar()
        import WorkingPlane

        WorkingPlane._view_observer_start()  # Updates the draftToolBar when switching views.
        from draftutils import grid_observer

        grid_observer._view_observer_setup()
        FreeCAD.Console.PrintLog("Draft workbench activated.\n")

    def Deactivated(self):
        """When quitting the workbench."""
        if hasattr(FreeCADGui, "draftToolBar"):
            FreeCADGui.draftToolBar.Deactivated()
        if hasattr(FreeCADGui, "Snapper"):
            FreeCADGui.Snapper.hide()
            from draftutils import init_draft_statusbar

            init_draft_statusbar.hide_draft_statusbar()
        import WorkingPlane

        WorkingPlane._view_observer_stop()
        from draftutils import grid_observer

        grid_observer._view_observer_setup()
        FreeCAD.Console.PrintLog("Draft workbench deactivated.\n")

    def ContextMenu(self, recipient):
        """Define an optional custom context menu."""
        has_text = False
        for o in FreeCADGui.Selection.getCompleteSelection():
            if hasattr(o.Object, "Text"):
                has_text = True
                break

        if has_text:
            import sys
            from draftguitools import gui_hyperlink

            hyperlinks_search = gui_hyperlink.Draft_Hyperlink()
            if hyperlinks_search.has_hyperlinks() and sys.platform in [
                "win32",
                "cygwin",
                "darwin",
                "linux",
            ]:
                self.appendContextMenu("", ["Draft_Hyperlink"])

        self.appendContextMenu("Utilities", self.context_commands)

    def GetClassName(self):
        """Type of workbench."""
        return "Gui::PythonWorkbench"


FreeCADGui.addWorkbench(DraftWorkbench)

# Preference pages for importing and exporting various file formats
# are independent of the loading of the workbench and can be loaded at startup
import Draft_rc
from PySide.QtCore import QT_TRANSLATE_NOOP

FreeCADGui.addPreferencePage(
    ":/ui/preferences-dxf.ui", QT_TRANSLATE_NOOP("QObject", "Import-Export")
)
FreeCADGui.addPreferencePage(
    ":/ui/preferences-dwg.ui", QT_TRANSLATE_NOOP("QObject", "Import-Export")
)
FreeCADGui.addPreferencePage(
    ":/ui/preferences-svg.ui", QT_TRANSLATE_NOOP("QObject", "Import-Export")
)
FreeCADGui.addPreferencePage(
    ":/ui/preferences-oca.ui", QT_TRANSLATE_NOOP("QObject", "Import-Export")
)

FreeCAD.__unit_test__ += ["TestDraftGui"]



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2011 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD Profile"
__author__ = "Yorik van Havre"
__url__ = "https://www.freecad.org"

## @package ArchProfile
#  \ingroup ARCH
#  \brief Profile tools for ArchStructure
#
#  This module provides tools to build base profiles
#  for Arch Structure elements

import csv
import os

import FreeCAD
import Draft

from FreeCAD import Vector
from draftutils import params

if FreeCAD.GuiUp:
    from PySide import QtCore, QtGui
    from PySide.QtCore import QT_TRANSLATE_NOOP
    import FreeCADGui
    from draftutils.translate import translate
else:
    # \cond
    def translate(ctxt, txt):
        return txt

    def QT_TRANSLATE_NOOP(ctxt, txt):
        return txt

    # \endcond


# Presets in the form: Class, Name, Profile type, [profile data]
# Search for profiles.csv in data/Mod/Arch/Presets and in the same folder as this file
# and in the user path
profilefiles = [
    os.path.join(FreeCAD.getResourceDir(), "Mod", "BIM", "Presets", "profiles.csv"),
    os.path.join(os.path.dirname(__file__), "Presets", "profiles.csv"),
    os.path.join(FreeCAD.getUserAppDataDir(), "BIM", "profiles.csv"),
]


def readPresets():

    Presets = []
    bid = 1  # Unique index
    for profilefile in profilefiles:
        if os.path.exists(profilefile):
            try:
                with open(profilefile, "r") as csvfile:
                    beamreader = csv.reader(csvfile)
                    for row in beamreader:
                        if (not row) or row[0].startswith("#"):
                            continue
                        try:
                            r = [bid, row[0], row[1], row[2]]
                            for i in range(3, len(row)):
                                r = r + [float(row[i])]
                            if not r in Presets:
                                Presets.append(r)
                            bid = bid + 1
                        except ValueError:
                            print("Skipping bad line: " + str(row))
            except IOError:
                print("Could not open ", profilefile)
    return Presets


class _Profile(Draft._DraftObject):
    """Superclass for Profile classes"""

    def __init__(self, obj, profile):
        self.Profile = profile
        Draft._DraftObject.__init__(self, obj, "Profile")

    def dumps(self):
        if hasattr(self, "Profile"):
            return self.Profile

    def loads(self, state):
        if isinstance(state, list):
            self.Profile = state
        self.Type = "Profile"

    def cleanProperties(self, obj):
        """Remove all Profile properties"""

        for prop in obj.PropertiesList:
            if obj.getGroupOfProperty(prop) == "Draft":
                obj.setPropertyStatus(prop, "-LockDynamic")
                obj.removeProperty(prop)


class _ProfileC(_Profile):
    """A parametric circular tubeprofile. Profile data: [Outside diameter, Wall thickness]"""

    def __init__(self, obj, profile):
        self.cleanProperties(obj)
        obj.addProperty(
            "App::PropertyLength",
            "OutDiameter",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "Outside Diameter"),
            locked=True,
        ).OutDiameter = profile[4]
        obj.addProperty(
            "App::PropertyLength",
            "Thickness",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "Wall thickness"),
            locked=True,
        ).Thickness = profile[5]
        _Profile.__init__(self, obj, profile)

    def execute(self, obj):
        import Part

        pl = obj.Placement
        c1 = Part.Circle()
        c1.Radius = obj.OutDiameter.Value / 2
        c2 = Part.Circle()
        c2.Radius = obj.OutDiameter.Value / 2 - obj.Thickness.Value
        cs1 = c1.toShape()
        cs2 = c2.toShape()
        p = Part.makeRuledSurface(cs2, cs1)
        p.reverse()
        obj.Shape = p
        obj.Placement = pl


class _ProfileH(_Profile):
    """A parametric H or I beam profile. Profile data: [width, height, web thickness, flange thickness] (see http://en.wikipedia.org/wiki/I-beam for reference)"""

    def __init__(self, obj, profile):
        self.cleanProperties(obj)
        obj.addProperty(
            "App::PropertyLength",
            "Width",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "Width of the beam"),
            locked=True,
        ).Width = profile[4]
        obj.addProperty(
            "App::PropertyLength",
            "Height",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "Height of the beam"),
            locked=True,
        ).Height = profile[5]
        obj.addProperty(
            "App::PropertyLength",
            "WebThickness",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "Thickness of the web"),
            locked=True,
        ).WebThickness = profile[6]
        obj.addProperty(
            "App::PropertyLength",
            "FlangeThickness",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "Thickness of the flanges"),
            locked=True,
        ).FlangeThickness = profile[7]
        _Profile.__init__(self, obj, profile)

    def execute(self, obj):
        import Part

        pl = obj.Placement
        p1 = Vector(-obj.Width.Value / 2, -obj.Height.Value / 2, 0)
        p2 = Vector(obj.Width.Value / 2, -obj.Height.Value / 2, 0)
        p3 = Vector(obj.Width.Value / 2, (-obj.Height.Value / 2) + obj.FlangeThickness.Value, 0)
        p4 = Vector(
            obj.WebThickness.Value / 2, (-obj.Height.Value / 2) + obj.FlangeThickness.Value, 0
        )
        p5 = Vector(obj.WebThickness.Value / 2, obj.Height.Value / 2 - obj.FlangeThickness.Value, 0)
        p6 = Vector(obj.Width.Value / 2, obj.Height.Value / 2 - obj.FlangeThickness.Value, 0)
        p7 = Vector(obj.Width.Value / 2, obj.Height.Value / 2, 0)
        p8 = Vector(-obj.Width.Value / 2, obj.Height.Value / 2, 0)
        p9 = Vector(-obj.Width.Value / 2, obj.Height.Value / 2 - obj.FlangeThickness.Value, 0)
        p10 = Vector(
            -obj.WebThickness.Value / 2, obj.Height.Value / 2 - obj.FlangeThickness.Value, 0
        )
        p11 = Vector(
            -obj.WebThickness.Value / 2, (-obj.Height.Value / 2) + obj.FlangeThickness.Value, 0
        )
        p12 = Vector(-obj.Width.Value / 2, (-obj.Height.Value / 2) + obj.FlangeThickness.Value, 0)
        p = Part.makePolygon([p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p1])
        p = Part.Face(p)
        # p.reverse()
        obj.Shape = p
        obj.Placement = pl


class _ProfileR(_Profile):
    """A parametric rectangular beam profile based on [Width, Height]"""

    def __init__(self, obj, profile):
        self.cleanProperties(obj)
        obj.addProperty(
            "App::PropertyLength",
            "Width",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "Width of the beam"),
            locked=True,
        ).Width = profile[4]
        obj.addProperty(
            "App::PropertyLength",
            "Height",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "Height of the beam"),
            locked=True,
        ).Height = profile[5]
        _Profile.__init__(self, obj, profile)

    def execute(self, obj):
        import Part

        pl = obj.Placement
        p1 = Vector(-obj.Width.Value / 2, -obj.Height.Value / 2, 0)
        p2 = Vector(obj.Width.Value / 2, -obj.Height.Value / 2, 0)
        p3 = Vector(obj.Width.Value / 2, obj.Height.Value / 2, 0)
        p4 = Vector(-obj.Width.Value / 2, obj.Height.Value / 2, 0)
        p = Part.makePolygon([p1, p2, p3, p4, p1])
        p = Part.Face(p)
        # p.reverse()
        obj.Shape = p
        obj.Placement = pl


class _ProfileRH(_Profile):
    """A parametric Rectangular hollow beam profile. Profile data: [width, height, thickness]"""

    def __init__(self, obj, profile):
        self.cleanProperties(obj)
        obj.addProperty(
            "App::PropertyLength",
            "Width",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "Width of the beam"),
            locked=True,
        ).Width = profile[4]
        obj.addProperty(
            "App::PropertyLength",
            "Height",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "Height of the beam"),
            locked=True,
        ).Height = profile[5]
        obj.addProperty(
            "App::PropertyLength",
            "Thickness",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "Thickness of the sides"),
            locked=True,
        ).Thickness = profile[6]
        _Profile.__init__(self, obj, profile)

    def execute(self, obj):
        import Part

        pl = obj.Placement
        p1 = Vector(-obj.Width.Value / 2, -obj.Height.Value / 2, 0)
        p2 = Vector(obj.Width.Value / 2, -obj.Height.Value / 2, 0)
        p3 = Vector(obj.Width.Value / 2, obj.Height.Value / 2, 0)
        p4 = Vector(-obj.Width.Value / 2, obj.Height.Value / 2, 0)
        q1 = Vector(
            -obj.Width.Value / 2 + obj.Thickness.Value,
            -obj.Height.Value / 2 + obj.Thickness.Value,
            0,
        )
        q2 = Vector(
            obj.Width.Value / 2 - obj.Thickness.Value,
            -obj.Height.Value / 2 + obj.Thickness.Value,
            0,
        )
        q3 = Vector(
            obj.Width.Value / 2 - obj.Thickness.Value, obj.Height.Value / 2 - obj.Thickness.Value, 0
        )
        q4 = Vector(
            -obj.Width.Value / 2 + obj.Thickness.Value,
            obj.Height.Value / 2 - obj.Thickness.Value,
            0,
        )
        p = Part.makePolygon([p1, p2, p3, p4, p1])
        q = Part.makePolygon([q1, q2, q3, q4, q1])
        # r = Part.Face([p,q])
        # r.reverse()
        p = Part.Face(p)
        q = Part.Face(q)
        r = p.cut(q)
        obj.Shape = r
        obj.Placement = pl


class _ProfileU(_Profile):
    """A parametric U profile. Profile data: [width, height, web thickness, flange thickness] (see  https://en.wikipedia.org/wiki/Structural_channel for reference)"""

    def __init__(self, obj, profile):
        self.cleanProperties(obj)
        obj.addProperty(
            "App::PropertyLength",
            "Width",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "Width of the beam"),
            locked=True,
        ).Width = profile[4]
        obj.addProperty(
            "App::PropertyLength",
            "Height",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "Height of the beam"),
            locked=True,
        ).Height = profile[5]
        obj.addProperty(
            "App::PropertyLength",
            "WebThickness",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "Thickness of the webs"),
            locked=True,
        ).WebThickness = profile[6]
        obj.addProperty(
            "App::PropertyLength",
            "FlangeThickness",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "Thickness of the flange"),
            locked=True,
        ).FlangeThickness = profile[7]
        _Profile.__init__(self, obj, profile)

    def execute(self, obj):
        import Part

        pl = obj.Placement
        p1 = Vector(-obj.Width.Value / 2, -obj.Height.Value / 2, 0)
        p2 = Vector(obj.Width.Value / 2, -obj.Height.Value / 2, 0)
        p3 = Vector(obj.Width.Value / 2, obj.Height.Value / 2, 0)
        p4 = Vector(obj.Width.Value / 2 - obj.FlangeThickness.Value, obj.Height.Value / 2, 0)
        p5 = Vector(
            obj.Width.Value / 2 - obj.FlangeThickness.Value,
            obj.WebThickness.Value - obj.Height.Value / 2,
            0,
        )
        p6 = Vector(
            -obj.Width.Value / 2 + obj.FlangeThickness.Value,
            obj.WebThickness.Value - obj.Height.Value / 2,
            0,
        )
        p7 = Vector(-obj.Width.Value / 2 + obj.FlangeThickness.Value, obj.Height.Value / 2, 0)
        p8 = Vector(-obj.Width.Value / 2, obj.Height.Value / 2, 0)
        p = Part.makePolygon([p1, p2, p3, p4, p5, p6, p7, p8, p1])
        p = Part.Face(p)
        # p.reverse()
        obj.Shape = p
        obj.Placement = pl


class _ProfileL(_Profile):
    """A parametric L profile. Profile data: [width, height, thickness]"""

    def __init__(self, obj, profile):
        self.cleanProperties(obj)
        obj.addProperty(
            "App::PropertyLength",
            "Width",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "Width of the beam"),
            locked=True,
        ).Width = profile[4]
        obj.addProperty(
            "App::PropertyLength",
            "Height",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "Height of the beam"),
            locked=True,
        ).Height = profile[5]
        obj.addProperty(
            "App::PropertyLength",
            "Thickness",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "Thickness of the legs"),
            locked=True,
        ).Thickness = profile[6]
        _Profile.__init__(self, obj, profile)

    def execute(self, obj):
        import Part

        pl = obj.Placement
        p1 = Vector(-obj.Width.Value / 2, obj.Height.Value / 2, 0)
        p2 = Vector(-obj.Width.Value / 2, -obj.Height.Value / 2, 0)
        p3 = Vector(obj.Width.Value / 2, -obj.Height.Value / 2, 0)
        p4 = Vector(obj.Width.Value / 2, -obj.Height.Value / 2 + obj.Thickness.Value, 0)
        p5 = Vector(
            -obj.Width.Value / 2 + obj.Thickness.Value,
            -obj.Height.Value / 2 + obj.Thickness.Value,
            0,
        )
        p6 = Vector(-obj.Width.Value / 2 + obj.Thickness.Value, obj.Height.Value / 2, 0)
        p = Part.makePolygon([p1, p2, p3, p4, p5, p6, p1])
        p = Part.Face(p)
        # p.reverse()
        obj.Shape = p
        obj.Placement = pl


class _ProfileT(_Profile):
    """A parametric T profile. Profile data: [width, height, web thickness, flange thickness]"""

    def __init__(self, obj, profile):
        self.cleanProperties(obj)
        obj.addProperty(
            "App::PropertyLength",
            "Width",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "Width of the beam"),
            locked=True,
        ).Width = profile[4]
        obj.addProperty(
            "App::PropertyLength",
            "Height",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "Height of the beam"),
            locked=True,
        ).Height = profile[5]
        obj.addProperty(
            "App::PropertyLength",
            "WebThickness",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "Thickness of the web"),
            locked=True,
        ).WebThickness = profile[6]
        obj.addProperty(
            "App::PropertyLength",
            "FlangeThickness",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "Thickness of the flanges"),
            locked=True,
        ).FlangeThickness = profile[7]
        _Profile.__init__(self, obj, profile)

    def execute(self, obj):
        import Part

        pl = obj.Placement
        p1 = Vector(obj.WebThickness.Value / 2, -obj.Height.Value / 2, 0)
        p2 = Vector(obj.WebThickness.Value / 2, obj.Height.Value / 2 - obj.FlangeThickness.Value, 0)
        p3 = Vector(obj.Width.Value / 2, obj.Height.Value / 2 - obj.FlangeThickness.Value, 0)
        p4 = Vector(obj.Width.Value / 2, obj.Height.Value / 2, 0)
        p5 = Vector(-obj.Width.Value / 2, obj.Height.Value / 2, 0)
        p6 = Vector(-obj.Width.Value / 2, obj.Height.Value / 2 - obj.FlangeThickness.Value, 0)
        p7 = Vector(
            -obj.WebThickness.Value / 2, obj.Height.Value / 2 - obj.FlangeThickness.Value, 0
        )
        p8 = Vector(-obj.WebThickness.Value / 2, -obj.Height.Value / 2, 0)
        p = Part.makePolygon([p1, p2, p3, p4, p5, p6, p7, p8, p1])
        p = Part.Face(p)
        # p.reverse()
        obj.Shape = p
        obj.Placement = pl


class _ProfileTSLOT(_Profile):
    """T slot profile common made of aluminum"""

    def __init__(self, obj, profile):
        self.cleanProperties(obj)
        obj.addProperty(
            "App::PropertyLength",
            "Size",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "Overall size"),
            locked=True,
        ).Size = profile[4]
        obj.addProperty(
            "App::PropertyLength",
            "SlotSize",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "Slot size"),
            locked=True,
        ).SlotSize = profile[5]
        obj.addProperty(
            "App::PropertyLength",
            "WallThickness",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "Thickness of the wall"),
            locked=True,
        ).WallThickness = profile[6]
        obj.addProperty(
            "App::PropertyLength",
            "TnutSlotWidth",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "T-nut slot width"),
            locked=True,
        ).TnutSlotWidth = profile[7]
        obj.addProperty(
            "App::PropertyLength",
            "TnutSlotDepth",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "T-nut slot depth"),
            locked=True,
        ).TnutSlotDepth = profile[8]
        obj.addProperty(
            "App::PropertyLength",
            "CoreSize",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "Internal core size"),
            locked=True,
        ).CoreSize = profile[9]
        obj.addProperty(
            "App::PropertyLength",
            "HoleDiameter",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "Internal hole diameter"),
            locked=True,
        ).HoleDiameter = profile[10]
        obj.addProperty(
            "App::PropertyLength",
            "FilletRadius",
            "Draft",
            QT_TRANSLATE_NOOP("App::Property", "Corner fillet radius"),
            locked=True,
        ).FilletRadius = profile[11]
        _Profile.__init__(self, obj, profile)

    def execute(self, obj):
        import Part
        import math

        # import DraftGeomUtils
        pl = obj.Placement
        nut_depth = obj.CoreSize / 2 - (obj.WallThickness / math.sqrt(2))
        core_dim = obj.CoreSize / 2
        wall_dim = obj.WallThickness
        slot_dim = obj.SlotSize / 2
        nut_width_dim = obj.TnutSlotWidth / 2
        templist = list()
        templist.append(Vector(nut_depth, core_dim, 0.0))
        templist.append(Vector(nut_width_dim, nut_width_dim + (wall_dim / math.sqrt(2)), 0.0))
        templist.append(Vector(nut_width_dim, obj.Size / 2 - wall_dim, 0.0))
        templist.append(Vector(slot_dim, obj.Size / 2 - wall_dim, 0.0))
        templist.append(Vector(slot_dim, obj.Size / 2, 0.0))
        templist.append(Vector(obj.Size / 2, obj.Size / 2, 0.0))
        # TODO:
        # add fillet four profile corners
        # sub_poly1 = Part.makePolygon(templist)
        # rd1=Vector(obj.Size/2-(obj.FilletRadius),obj.Size,0)
        # rd2=Vector(obj.Size/2-(obj.FilletRadius-obj.FilletRadius*math.sin(math.pi/4)),obj.Size/2-(obj.FilletRadius-obj.FilletRadius*math.sin(math.pi/4)),0)
        # rd3=Vector(obj.Size,obj.Size/2-(obj.FilletRadius),0)
        # templist.append(Part.Arc(rd1,rd2,rd3))
        templist.append(Vector(obj.Size / 2, slot_dim, 0.0))
        templist.append(Vector(obj.Size / 2 - wall_dim, slot_dim, 0.0))
        templist.append(Vector(obj.Size / 2 - wall_dim, nut_width_dim, 0.0))
        templist.append(Vector(nut_width_dim + (wall_dim / math.sqrt(2)), nut_width_dim, 0.0))
        templist.append(Vector(core_dim, nut_depth, 0.0))
        # sub_poly2 = Part.makePolygon(templist3)
        # l1=Part.LineSegment(Vector(obj.SlotSize/2, obj.Size/2, 0.0),Vector(obj.Size/2, obj.Size/2, 0.0))
        # l2=Part.LineSegment(Vector(obj.Size/2, obj.Size/2, 0.0),Vector(obj.Size/2, 2.7500000000022977, 0.0))
        # round=DraftGeomUtils.fillet([l2.toShape(),l1.toShape()],obj.FilletRadius)
        # wire=Part.Wire(round)
        # p = Part.Shape([sub_poly1,round,sub_poly2])
        # p=Part.Face(Shape.Edges)
        templist2 = list()
        for vec in reversed(templist):
            templist2.append(Vector(vec.x, vec.y * -1, vec.z))
        templist = templist + templist2
        templist2.clear()
        for vec in reversed(templist):
            templist2.append(Vector(vec.x * -1, vec.y, vec.z))
        templist = templist + templist2
        templist.append(templist[0])
        poly = Part.makePolygon(templist)
        pf = Part.Face(poly)
        hole = Part.makeCircle(obj.HoleDiameter / 2, FreeCAD.Vector(0, 0, 0))
        cf = Part.Face(Part.Wire(hole))
        p = pf.cut(cf)
        # p.reverse()
        obj.Shape = p
        obj.Placement = pl


class ViewProviderProfile(Draft._ViewProviderDraft):
    """General view provider for Profile classes"""

    def __init__(self, vobj):

        Draft._ViewProviderDraft.__init__(self, vobj)

    def getIcon(self):

        import Arch_rc

        return ":/icons/Arch_Profile.svg"

    def setEdit(self, vobj, mode):
        if mode == 1 or mode == 2:
            return None

        taskd = ProfileTaskPanel(vobj.Object)
        FreeCADGui.Control.showDialog(taskd)
        return True

    def unsetEdit(self, vobj, mode):
        if mode == 1 or mode == 2:
            return None

        FreeCADGui.Control.closeDialog()
        return True


class ProfileTaskPanel:
    """The editmode TaskPanel for Profile objects"""

    def __init__(self, obj):

        self.obj = obj
        self.Profile = None
        if hasattr(obj.Proxy, "Profile"):
            self.Profile = obj.Proxy.Profile
        if isinstance(self.obj.Proxy, _ProfileC):
            self.type = "C"
        elif isinstance(self.obj.Proxy, _ProfileH):
            self.type = "H"
        elif isinstance(self.obj.Proxy, _ProfileR):
            self.type = "R"
        elif isinstance(self.obj.Proxy, _ProfileRH):
            self.type = "RH"
        elif isinstance(self.obj.Proxy, _ProfileU):
            self.type = "U"
        elif isinstance(self.obj.Proxy, _ProfileL):
            self.type = "L"
        elif isinstance(self.obj.Proxy, _ProfileT):
            self.type = "T"
        elif isinstance(self.obj.Proxy, _ProfileTSLOT):
            self.type = "TSLOT"
        else:
            self.type = "Building Element Proxy"
        self.form = QtGui.QWidget()
        layout = QtGui.QVBoxLayout(self.form)
        self.comboCategory = QtGui.QComboBox(self.form)
        layout.addWidget(self.comboCategory)
        self.comboProfile = QtGui.QComboBox(self.form)
        layout.addWidget(self.comboProfile)
        QtCore.QObject.connect(
            self.comboCategory, QtCore.SIGNAL("currentTextChanged(QString)"), self.changeCategory
        )
        QtCore.QObject.connect(
            self.comboProfile, QtCore.SIGNAL("currentIndexChanged(int)"), self.changeProfile
        )
        # Read preset profiles and add relevant ones
        self.categories = []
        self.presets = readPresets()
        for pre in self.presets:
            if pre[1] not in self.categories:
                self.categories.append(pre[1])
        self.comboCategory.addItem(" ")
        if self.categories:
            self.comboCategory.addItems(self.categories)
        # Find current profile by label
        if not self.Profile:
            for pre in self.presets:
                if self.type == pre[1]:
                    if self.obj.Label in pre[2]:
                        self.Profile = pre
                        break
        if not self.Profile:
            # try to find by size
            if hasattr(self.obj, "Width") and hasattr(self.obj, "Height"):
                for pre in self.presets:
                    if self.type == pre[1]:
                        if (
                            abs(self.obj.Width - self.Profile[4]) < 0.1
                            and abs(self.obj.Height - self.Profile[5]) < 0.1
                        ):
                            self.Profile = pre
                            break
        if self.Profile:
            origprofile = list(self.Profile)  # the operation below will change self.profile
            self.comboCategory.setCurrentIndex(1 + self.categories.index(origprofile[1]))
            self.changeCategory(origprofile[1])
            self.comboProfile.setCurrentIndex(self.currentpresets.index(origprofile))
        self.retranslateUi(self.form)

    def changeCategory(self, text):

        self.comboProfile.clear()
        self.currentpresets = []
        for pre in self.presets:
            if pre[1] == text:
                self.currentpresets.append(pre)
                f = FreeCAD.Units.Quantity(pre[4], FreeCAD.Units.Length).getUserPreferred()
                d = params.get_param("Decimals", path="Units")
                s1 = str(round(pre[4] / f[1], d))
                s2 = str(round(pre[5] / f[1], d))
                s3 = str(f[2])
                self.comboProfile.addItem(pre[2] + " (" + s1 + "x" + s2 + s3 + ")")

    def changeProfile(self, idx):

        self.Profile = self.currentpresets[idx]

    def accept(self):

        self.obj.Label = self.Profile[2] + "_"
        if self.Profile:
            if self.Profile[3] == "C":
                _ProfileC(self.obj, self.Profile)
            elif self.Profile[3] == "H":
                _ProfileH(self.obj, self.Profile)
            elif self.Profile[3] == "R":
                _ProfileR(self.obj, self.Profile)
            elif self.Profile[3] == "RH":
                _ProfileRH(self.obj, self.Profile)
            elif self.Profile[3] == "U":
                _ProfileU(self.obj, self.Profile)
            elif self.Profile[3] == "L":
                _ProfileL(self.obj, self.Profile)
            elif self.Profile[3] == "T":
                _ProfileT(self.obj, self.Profile)
            elif self.Profile[3] == "TSLOT":
                _ProfileTSLOT(self.obj, self.Profile)
            else:
                print("Profile not supported")

            FreeCAD.ActiveDocument.recompute()
            FreeCADGui.ActiveDocument.resetEdit()
        return True

    def reject(self):

        FreeCADGui.ActiveDocument.resetEdit()
        return True

    def retranslateUi(self, TaskPanel):

        self.form.setWindowTitle(
            self.type + " " + QtGui.QApplication.translate("Arch", "Profile", None)
        )



# ***************************************************************************
# *   Copyright (c) 2016 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM mesh gmsh task panel for the document object"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## @package task_mesh_gmsh
#  \ingroup FEM
#  \brief task panel for mesh gmsh object

from PySide import QtCore

import FreeCAD
import FreeCADGui

from femmesh import gmshtools

from . import base_femlogtaskpanel


class _TaskPanel(base_femlogtaskpanel._BaseLogTaskPanel):
    """
    The TaskPanel for editing References property of
    MeshGmsh objects and creation of new FEM mesh
    """

    def __init__(self, obj):
        super().__init__(obj, gmshtools.GmshTools(obj))

        self.form = FreeCADGui.PySideUic.loadUi(
            FreeCAD.getHomePath() + "Mod/Fem/Resources/ui/MeshGmsh.ui"
        )
        self.text_log = self.form.te_output
        self.text_time = self.form.l_time

        self.setup_connections()

    def setup_connections(self):
        super().setup_connections()

        QtCore.QObject.connect(
            self.form.qsb_max_size, QtCore.SIGNAL("valueChanged(Base::Quantity)"), self.max_changed
        )
        QtCore.QObject.connect(
            self.form.qsb_min_size, QtCore.SIGNAL("valueChanged(Base::Quantity)"), self.min_changed
        )
        QtCore.QObject.connect(
            self.form.cb_dimension, QtCore.SIGNAL("activated(int)"), self.choose_dimension
        )
        QtCore.QObject.connect(
            self.form.cb_order, QtCore.SIGNAL("activated(int)"), self.choose_order
        )
        self.form.cb_dimension.addItems(self.obj.getEnumerationsOfProperty("ElementDimension"))

        self.form.cb_order.addItems(self.obj.getEnumerationsOfProperty("ElementOrder"))
        QtCore.QObject.connect(
            self.form.pb_get_gmsh_version, QtCore.SIGNAL("clicked()"), self.get_version
        )

        self.get_object_params()
        self.set_widgets()

    def get_object_params(self):
        self.clmax = self.obj.CharacteristicLengthMax
        self.clmin = self.obj.CharacteristicLengthMin
        self.dimension = self.obj.ElementDimension
        self.order = self.obj.ElementOrder

    def set_object_params(self):
        self.obj.CharacteristicLengthMax = self.clmax
        self.obj.CharacteristicLengthMin = self.clmin
        self.obj.ElementDimension = self.dimension
        self.obj.ElementOrder = self.order

    def set_widgets(self):
        "fills the widgets"
        self.form.qsb_max_size.setProperty("value", self.clmax)
        FreeCADGui.ExpressionBinding(self.form.qsb_max_size).bind(
            self.obj, "CharacteristicLengthMax"
        )
        self.form.qsb_min_size.setProperty("value", self.clmin)
        FreeCADGui.ExpressionBinding(self.form.qsb_min_size).bind(
            self.obj, "CharacteristicLengthMin"
        )
        index_dimension = self.form.cb_dimension.findText(self.dimension)
        self.form.cb_dimension.setCurrentIndex(index_dimension)
        index_order = self.form.cb_order.findText(self.order)
        self.form.cb_order.setCurrentIndex(index_order)

    def max_changed(self, base_quantity_value):
        self.clmax = base_quantity_value

    def min_changed(self, base_quantity_value):
        self.clmin = base_quantity_value

    def choose_dimension(self, index):
        if index < 0:
            return
        self.form.cb_dimension.setCurrentIndex(index)
        self.dimension = self.form.cb_dimension.itemText(index)

    def choose_order(self, index):
        if index < 0:
            return
        self.form.cb_order.setCurrentIndex(index)
        self.order = self.form.cb_order.itemText(index)
